(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Relations.

Require Import utils.

Set Implicit Arguments.

Local Notation crt := (clos_refl_trans _).

Definition square X A B Y (R : rel X A) (S : rel X B) (T : rel A Y) (U : rel B Y) :=
  forall x a b, R x a -> S x b -> exists y, T a y /\ U b y.

Fact square_Acc X Y (R : rel X X) (S : rel Y Y) (E : rel X Y) :
  square R E E S -> forall x y, E x y -> Acc (inv S) y -> Acc (inv R) x.
Proof.
  intros H0 x y H Hy.
  revert Hy x H.
  induction 1 as [ y Hy IH ]; intros x H.
  constructor 1; intros x' Hx.
  destruct (H0 _ _ _ Hx H) as (y' & H1 & H2).
  apply (IH _ H2 _ H1).
Qed. 

Section squares.

  Fact square_comm X A B Y R S T U : @square X A B Y R S T U -> square S R U T.
  Proof. 
    intros H x a b H1 H2.
    destruct (H _ _ _ H2 H1) 
      as (y & ? ); exists y; tauto.
  Qed.
  
  Fact square_eq X Y (S : rel X Y) : square eq S S eq.
  Proof.
    intros x a b ? ?; subst a; exists b; auto.
  Qed. 

  Fact square_dec X A B Y (R : rel X A) S T U (R' : rel X A) : 
     R' inc2 R -> @square X A B Y R S T U -> square R' S T U.
  Proof.
    intros I H x a b H1 H2.
    destruct (H _ _ _ (I _ _ H1) H2) 
      as (y & ? & ?); exists y; split; auto.
  Qed.

  Fact square_inc X A B Y R S (T : rel A Y) U (T' : rel A Y) : 
     T inc2 T' -> @square X A B Y R S T U -> square R S T' U.
  Proof.
    intros ? H x a b H1 H2.
    destruct (H _ _ _ H1 H2) 
      as (y & ? & ?); exists y; split; auto.
  Qed.

  Fact square_rt X Y R S T : 
      @square X X Y Y R S S T -> square (crt R) S S (crt T).
  Proof.
    intros H x1 x2 y H1; revert H1 y.
    induction 1 as [ x1 x2 H12 | x1 | x1 x2 x3 H1 IH1 H2 IH2 ]; intros y1 Hy1.

    destruct (H _ _ _ H12 Hy1) as (y2 & ? & ?).
    exists y2; split; auto; constructor 1; auto.
 
    exists y1; split; auto; constructor 2.

    destruct (IH1 _ Hy1) as (y2 & Hy2 & ?).
    destruct (IH2 _ Hy2) as (y3 & Hy3 & ?).
    exists y3; split; auto.
    constructor 3 with y2; auto.
  Qed.
  
  Fact square_cup X A B Y R1 R2 S T U:
       square R1 S T U
    -> square R2 S T U
    -> @square X A B Y (R1 cup2 R2) S T U.
  Proof.
    intros H1 H2 x ? ? [] ?.
    apply H1 with x; auto.
    apply H2 with x; auto.
  Qed.

  Fact square_inc_eq X (R S T : rel X X) : S inc2 T
                            -> square R S T (R cup2 eq)
                            -> square (R cup2 eq) S T (R cup2 eq).
  Proof.
    intros H1 H2 x a b [ Ha | [] ] Hb.
    destruct (H2 _ _ _ Ha Hb) as (y & ? & ?); exists y; split; auto.
    exists b; auto.
  Qed.

  Fact crt_eq_inc X (R : rel X X) : crt (R cup2 eq) inc2 crt R.
  Proof.
    intros x y H; apply clos_refl_trans_idem; revert x y H.
    apply clos_refl_trans_mono.
    intros ? ? [|[]].
    constructor 1; auto.
    constructor 2.
  Qed.

  Fact crt_inc_eq X (R : rel X X) : crt R inc2 crt (R cup2 eq).
  Proof.
    apply clos_refl_trans_mono; left; auto.
  Qed.

  Fact crt_inv_inc X (R : rel X X) : crt (inv R) inc2 inv (crt R).
  Proof.
    induction 1 as [ | | ? y ].
    constructor 1; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.

  Fact inv_crt_inc X (R : rel X X) : inv (crt R) inc2 crt (inv R).
  Proof.
    induction 1 as [ | | ? y ].
    constructor 1; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.
  
  Fact crt_inv_eq X (R S : rel X X) : (forall x y, crt R x y <-> crt S x y) -> (forall x y, crt (inv R) x y <-> crt (inv S) x y).
  Proof.
    intros H ? ?; split; intro;
    apply crt_inv_inc, H, inv_crt_inc; auto.
  Qed.
 
  Fact square_prop5 X (R S : rel X X) : 
                            square      R       S  (crt S) (R cup2 eq)
                         -> square (crt R) (crt S) (crt S) (crt R).
  Proof.
    simpl; intros H.
    apply square_comm, square_inc with (1 := @crt_eq_inc _ _),
          square_comm, square_dec with (1 := @crt_inc_eq _ _),
          square_rt, square_inc with (1 := @clos_refl_trans_idem _ _),
          square_comm, square_rt, 
          square_comm, square_inc_eq; auto.
    apply clos_refl_trans_inc.
  Qed.

  Fact square_crt X (R S : rel X X) : 
     square R S S R -> square (crt R) (crt S) (crt S) (crt R).
  Proof.
    intro; do 2 (apply square_rt, square_comm); auto. 
  Qed.

  Definition commute X (R S : rel X X) := square R S S R.
  Definition diamond X (R : rel X X) := square R R R R.
  Definition confluent X (R : rel X X) := diamond (crt R).

  Fact square_prop6 X (R S : rel X X) : 
     S inc2 R -> R inc2 crt S -> diamond R -> confluent S.
  Proof.
    intros H1 H2 H3; red.
    assert (crt R inc2 crt S) as H4.
      apply inc2_trans with (2 := @clos_refl_trans_idem _ _), 
            clos_refl_trans_mono; auto.
    assert (crt S inc2 crt R) as H5.
      apply clos_refl_trans_mono; auto.
    do 2 apply square_inc with (1 := H4), square_comm.
    do 2 apply square_dec with (1 := H5), square_comm.
    apply square_crt; auto.
  Qed.

  Fact square_prop7 X (R : rel X X) : square R R (R cup2 eq) (R cup2 eq)
                                   -> confluent R.
  Proof.
    intros H.
    apply square_prop5, square_inc with (2 := H).
    intros ? ? [ | []].
    constructor 1; auto.
    constructor 2.
  Qed.
  
  Section Hindley_Rosen.
  
    Variables (X : Type) (R S : rel X X).
    
    Hypothesis HR  : confluent R.
    Hypothesis HS  : confluent S.
    Hypothesis HRS : commute (crt R) (crt S).
    
    Let lemma1 : square (crt (R cup2 S)) (crt R) (crt R) (crt (R cup2 S)).
    Proof.
      apply square_comm, square_inc with (1 := @clos_refl_trans_idem _ _).
      apply square_comm, square_rt.
      intros x a b [ Ha | Ha ] Hb.
      
      destruct (@HR x a b) as (y & Hy1 & Hy2); auto.
      constructor 1; auto.
      exists y; split; auto.
      revert Hy2; apply clos_refl_trans_mono; auto.
      
      destruct (@HRS x b a) as (y & Hy1 & Hy2); auto.
      constructor 1; auto.
      exists y; split; auto.
      revert Hy1; apply clos_refl_trans_mono; auto.
    Qed.
    
    Let lemma2 : square (crt (R cup2 S)) (crt S) (crt S) (crt (R cup2 S)).
    Proof.
      apply square_comm, square_inc with (1 := @clos_refl_trans_idem _ _).
      apply square_comm, square_rt.
      intros x a b [ Ha | Ha ] Hb.
      
      destruct (@HRS x a b) as (y & Hy1 & Hy2); auto.
      constructor 1; auto.
      exists y; split; auto.
      revert Hy2; apply clos_refl_trans_mono; auto.
      
      destruct (@HS x a b) as (y & Hy1 & Hy2); auto.
      constructor 1; auto.
      exists y; split; auto.
      revert Hy2; apply clos_refl_trans_mono; auto.     
    Qed.
    
    Let lemma3 : square (crt (R cup2 S)) R (crt (R cup2 S)) (crt (R cup2 S)).
    Proof.
      apply square_inc with (crt R).
      apply clos_refl_trans_mono; auto.
      apply square_comm, square_dec with (1 := @clos_refl_trans_inc _ _).
      apply square_comm; auto.
    Qed.
    
    Let lemma4 : square (crt (R cup2 S)) S (crt (R cup2 S)) (crt (R cup2 S)).
    Proof.
      apply square_inc with (crt S).
      apply clos_refl_trans_mono; auto.
      apply square_comm, square_dec with (1 := @clos_refl_trans_inc _ _).
      apply square_comm; auto.
    Qed.
    
    Lemma Hindley_Rosen : confluent (R cup2 S).
    Proof.
      apply square_comm, square_inc with (1 := @clos_refl_trans_idem _ _), 
            square_comm, square_rt, square_cup;
      apply square_comm; auto.
    Qed.
    
  End Hindley_Rosen.
  
  Fact diamond_eq X (R S : rel X X) : (forall x y, R x y <-> S x y) -> diamond R -> diamond S.
  Proof.
    intros E H x a b; do 2 rewrite <- E; intros Ha Hb.
    destruct (H _ _ _ Ha Hb) as (y & ? & ?).
    exists y; do 2 rewrite <- E; auto.
  Qed.
  
  Fact confluent_eq X (R S : rel X X) : (forall x y, R x y <-> S x y) -> confluent R -> confluent S.
  Proof.
    intros H; apply diamond_eq.
    split; apply clos_refl_trans_mono; intros ? ?; apply H.
  Qed.
  
End squares.

Definition church_rosser X (R : rel X X) := diamond R.

Fact cr_eq X R S : (forall x y, R x y <-> S x y) -> @church_rosser X S -> church_rosser R.
Proof.
  intros E; apply diamond_eq.
  split; apply E.
Qed.

Theorem cr_clos_refl_trans X R : @church_rosser X R -> church_rosser (crt R).
Proof.
  apply square_crt.
Qed.
     
Theorem cr_clos_refl X (R : X -> X -> Prop) : 
      (forall x a b, R x a -> R x b -> a = b \/ exists y, R a y /\ R b y)
   -> church_rosser (crt R).
Proof.
  intros.
  set (R' x y := x = y \/ R x y).
  apply cr_eq with (S := clos_refl_trans _ R').
  intros x y; split.
  apply clos_refl_trans_mono; right; auto.
  apply inc2_trans with (2 := @clos_refl_trans_idem _ _), 
        clos_refl_trans_mono.
  intros ? ? [?|?].
  subst; constructor 2.
  constructor 1; auto.
  apply cr_clos_refl_trans.
  intros x a b; intros [ Ha | Ha ] [ Hb | Hb ]; subst; unfold R'.
  exists b; auto.
  exists b; auto.
  exists a; auto.
  destruct (H _ _ _ Ha Hb) as [ H1 | (y & H1 & H2) ].
  subst; exists b; auto.
  exists y; auto.
Qed.
  



  