(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Max Omega List Relations.

Require Import notations tac_utils rel_utils.

Set Implicit Arguments.

Fact map_eq_nil X Y (f : X -> Y) l : map f l = nil <-> l = nil.
Proof.
  split.
  destruct l; auto; discriminate 1.
  intros; subst; auto.
Qed.

Fact map_neq_nil X Y (f : X -> Y) l : map f l <> nil <-> l <> nil.
Proof.
  unfold not; rewrite map_eq_nil; tauto.
Qed.

Fact app_eq_nil X (l m : list X) : l ++ m = nil <-> l = nil /\ m = nil.
Proof.
  split.
  destruct l; destruct m; auto; discriminate 1.
  intros [? ?]; subst; auto.
Qed.
 
Fact app_neq_nil X (l m : list X) : l ++ m <> nil <-> l <> nil \/ m <> nil.
Proof.
  destruct l.
  destruct m.
  tauto.
  tauto.
  split.   
  left; discriminate.
  discriminate.
Qed.

Fact flat_map_neq_nil X Y (f : X -> list Y) l : 
    flat_map f l <> nil <-> exists x, In x l /\ f x <> nil.
Proof.
  split. 

  induction l as [ | x l ]; simpl.
  destruct 1; auto.
  intros H; apply app_neq_nil in H.
  destruct H as [ H | H ].
  exists x; tauto.
  destruct (IHl H) as (y & ? & ?).
  exists y; tauto.

  intros (x & Hx & Hf).
  induction l as [ | y l IH ]; destruct Hx.
  subst; simpl.
  apply app_neq_nil; tauto.
  simpl.
  apply app_neq_nil; right; auto.    
Qed.

Fact Forall_rev X P l : @Forall X P l <-> Forall P (rev l).
Proof.
  split; intros H; apply Forall_forall; intros x Hx.
  apply (proj1 (Forall_forall _ _) H), in_rev; auto.
  apply in_rev in Hx.
  apply (proj1 (Forall_forall _ _) H _ Hx).
Qed.

Fact Forall2_map X Y (f g : X -> Y) (R : Y -> Y -> Prop) l : Forall (fun x => R (f x) (g x)) l -> Forall2 R (map f l) (map g l).
Proof.
  induction 1; simpl; auto.
Qed.

Fact Forall_cons_inv X (P : X -> Prop) x l : Forall P (x::l) -> P x /\ Forall P l.
Proof.
  inversion_clear 1; tauto.
Qed. 

Section Forall2.

  Variable (X Y : Type). 

  Implicit Types (R S : X -> Y -> Prop).

  Fact exists_Forall2 R l : (forall x, In x l -> exists y, R x y)
                       -> exists m, Forall2 R l m.
  Proof.
    induction l as [ | x l IH ].
    exists nil; auto.
    intros H.
    destruct (H _ (or_introl eq_refl)) as (y & Hy).
    destruct IH as (m & Hm).
    intros; apply H; right; auto.
    exists (y::m); constructor; auto.
  Qed.

  Fact sig_Forall2 R l : (forall x, In x l -> { y | R x y }) 
                       -> { m | Forall2 R l m }.  
  Proof.
    induction l as [ | x l IHl ]; intros H.
    exists nil; auto.
    destruct IHl as (m & Hm).
    intros; apply H; right; auto.
    destruct (H x) as (y & Hy). 
    left; auto.
    exists (y::m); constructor; auto.
  Qed.

  Fact Forall2_left_inv R l m : Forall2 R l m -> forall x, In x l -> exists y, In y m /\ R x y.
  Proof.
    induction 1 as [ | a b l m H1 H2 IH2 ].
    destruct 1.
    intros z [ ? | ? ]; subst.
    exists b; split; auto; left; auto.
    destruct (IH2 _ H) as (y & ? & ?). 
    exists y; split; auto; right; auto.
  Qed. 

  Definition Forall_In_left_inv := Forall2_left_inv.

  Fact Forall2_right_inv R l m : Forall2 R l m -> forall y, In y m -> exists x, In x l /\ R x y.
  Proof.
    induction 1 as [ | a b l m H1 H2 IH2 ].
    destruct 1.
    intros z [ ? | ? ]; subst.
    exists a; split; auto; left; auto.
    destruct (IH2 _ H) as (x & ? & ?). 
    exists x; split; auto; right; auto.
  Qed.

  Fact Forall2_incl R S : (forall x y, R x y -> S x y) -> forall l m, Forall2 R l m -> Forall2 S l m.
  Proof.
    intros H; induction 1; constructor; auto.
  Qed.

  Fact Forall2_sym R l m : Forall2 R l m -> Forall2 (inv R) m l.
  Proof.
    induction 1; constructor; auto.
  Qed.

  Let Forall2_rev_app R l m a b : Forall2 R l m -> Forall2 R a b ->  Forall2 R (rev l++a) (rev m++b).
  Proof.
    intros H; revert H a b.
    induction 1; simpl; auto; intros a b Hab.
    do 2 rewrite <- app_assoc; simpl; auto.
  Qed.

  Fact Forall2_rev R l m : Forall2 R l m -> Forall2 R (rev l) (rev m).
  Proof.
    rewrite (app_nil_end (rev m)).
    rewrite (app_nil_end (rev l)).
    intros; apply Forall2_rev_app; auto.
  Qed.   

End Forall2.

Fact Forall2_map_right X Y Y' (R : X -> Y' -> Prop) (f : Y -> Y') l m : 
   Forall2 (fun x y => R x (f y)) l m <-> Forall2 R l (map f m).
Proof.
  split.
  induction 1; simpl; constructor; auto.
  gen_eq (map f m).
  intros m' H1 H2; revert H2 m H1.
  induction 1 as [ | x l y m' H1 H2 IH2 ]; intros m E; subst.
  destruct m.
  constructor.
  discriminate E.
  destruct m as [ | a m ].
  discriminate E.
  injc E.
  constructor; auto.
Qed.

Fact Forall2_with_refl X (R : rel X X) l : (Forall (fun x => R x x) l) -> Forall2 R l l.
Proof.
  induction 1; constructor; auto.
Qed.

Fact Forall2_comp X Y Z (R : rel X Y) (S : rel Y Z) l m k : Forall2 R l m -> Forall2 S m k -> Forall2 (fun x z => exists y, R x y /\ S y z) l k.
Proof.
  intros H; revert H k.
  induction 1; intros k.
  inversion_clear 1; constructor.
  inversion_clear 1; constructor; auto.
  exists y; auto.
Qed.

Fact clos_refl_trans_Forall2 X (R : X -> X -> Prop) l m : clos_refl_trans _ (Forall2 R) l m -> Forall2 (clos_refl_trans _ R) l m.
Proof.
  induction 1.
  revert H; apply Forall2_incl; intros; constructor 1; auto.
  apply Forall2_with_refl, Forall_forall; intros; constructor 2.
  generalize (Forall2_comp IHclos_refl_trans1 IHclos_refl_trans2).
  apply Forall2_incl; intros ? ? (w & ? & ?); constructor 3 with w; auto.
Qed.

Section Forall2_clos_refl_trans.
 
  Variables (X : Type) (R : X -> X -> Prop).

  Hypothesis R_refl : forall x, R x x.

  Let Forall2_refl l : Forall2 R l l.
  Proof.
    induction l; constructor; auto.
  Qed.

  Let crt_Forall_cons_1 l x y : clos_refl_trans _ R x y -> clos_refl_trans _ (Forall2 R) (x::l) (y::l).
  Proof.
    induction 1.
    constructor 1; constructor; auto.
    constructor 2.
    constructor 3 with (y::l); auto.
  Qed.

  Let crt_Forall_cons_2 x l m : clos_refl_trans _ (Forall2 R) l m -> clos_refl_trans _ (Forall2 R) (x::l) (x::m).
  Proof.
    induction 1.
    constructor 1; constructor; auto.
    constructor 2.
    constructor 3 with (x::y); auto.
  Qed.

  Fact Forall2_clos_refl_trans : Forall2 (clos_refl_trans _ R) inc2 clos_refl_trans _ (Forall2 R).
  Proof.
    induction 1.
    constructor 2.
    constructor 3 with (x::l'); auto.
  Qed.

End Forall2_clos_refl_trans.

Fact map_inj X Y (f : X -> Y) : (forall a b, f a = f b -> a = b) -> forall l m, map f l = map f m -> l = m.
Proof.
  intros Hf.
  induction l as [ | a l IHl ]; intros [ | b m ]; auto; try discriminate 1; simpl.
  intros H; injection H; clear H; intros H1 H2.
  apply Hf in H2; subst.
  apply IHl in H1; subst.
  auto.
Qed.

Fact map_id X l : map (fun x : X => x) l = l.
Proof.
  induction l; simpl; f_equal; auto.
Qed.

Fact rev_inj X (l m : list X) : rev l = rev m -> l = m.
Proof.
  intros H; rewrite <- (rev_involutive l), H, rev_involutive; trivial.
Qed.

Section list_an.

  Fixpoint list_an a n := 
    match n with 
      | 0   => nil
      | S n => a::list_an (S a) n
    end.

  Fact list_an_comp a n m : list_an a n ++ list_an (a+n) m = list_an a (n+m).
  Proof.
    revert a; induction n as [ | n IHn ]; intros a; simpl; auto.
    cutrewrite (a+S n = S a + n); try omega.
    f_equal; auto.
  Qed.

  Let list_an_prop a n b : a+n <= b -> map (fun x => b - S x) (list_an a n) = rev (list_an (b - (a+n)) n).
  Proof.
    revert a b; induction n; intros a b H; auto.
    simpl (list_an a (S n)); simpl map.
    rewrite IHn; try omega.
    apply rev_inj.
    rewrite rev_involutive.
    simpl rev.
    rewrite rev_involutive.
    cutrewrite (a+S n = S (a+n)); try omega.
    cutrewrite (S n = n+1); try omega.
    rewrite <- list_an_comp; simpl.
    f_equal; f_equal; omega.
  Qed.

  Definition list_n n := list_an 0 n.

  Let list_n_rev_r n : map (fun x => n - x) (list_n (S n)) = rev (list_n (S n)).
  Proof.
    unfold list_n.
    generalize (list_an_prop 0 (S n) (le_refl _)); intros H.
    simpl plus in H.
    cutrewrite (S n - S n = 0) in H; try omega.
    rewrite <- H.
    apply map_ext.
    intros; omega.
  Qed.

  Fact list_n_rev n : map (fun x => n - S x) (list_n n) = rev (list_n n).
  Proof.
    destruct n.
    simpl; auto.
    rewrite <- list_n_rev_r.
    apply map_ext; intro; omega.
  Qed.

End list_an.

Section list_In_map.

  Variables (X Y : Type).

  Fixpoint list_In_map (l : list X) : (forall x, In x l -> Y) -> list Y.
  Proof.
    destruct l as [ | x l ].
    intros _; exact nil.
    intros H.
    refine (H x _:: (list_In_map l _)).
    left; reflexivity.
    intros a ?; apply H with a; right; assumption.
  Defined.

  Fact list_In_map_length ll Hll : length (list_In_map ll Hll) = length ll.
  Proof.
    induction ll; simpl; f_equal; auto.
  Qed.

  Fact list_In_map_ext l f g : (forall x Hx, f x Hx = g x Hx) -> list_In_map l f = list_In_map l g.
  Proof.
    intro; induction l; simpl; f_equal; auto.
  Qed.

  Fact list_In_map_eq_map f ll : list_In_map ll (fun x _ => f x) = map f ll.
  Proof.
    induction ll; simpl; f_equal; assumption.
  Qed.

  Fact In_list_In_map l f x (Hx : In x l) : In (f x Hx) (list_In_map l f). 
  Proof.
    revert f x Hx; induction l; simpl; intros f x Hx.
    destruct Hx.
    destruct Hx as [ ? | Hx ].
    subst; left; auto.
    right. 
    apply IHl with (f := fun x Hx => f x (or_intror Hx)).
  Qed.

  Fact list_In_map_In l f y : In y (list_In_map l f) -> exists x Hx, y = f x Hx.
  Proof.
    revert f y; induction l as [ | x l IH ]; intros f y.
    intros [].
    intros [ ? | H ]; subst.
    exists x, (or_introl eq_refl); auto.
    destruct IH with (f := fun x Hx => f x (or_intror Hx))
                     (1 := H)
     as (u & Hu & ?).
    exists u, (or_intror Hu); auto.
  Qed.

  Hypothesis Yeq_dec : forall x y : Y, { x = y } + { x <> y }.

  Fact list_In_map_In_dec l f y : In y (list_In_map l f) -> {x : _ & { Hx | y = f x Hx } }.
  Proof.
    revert f y; induction l as [ | x l IH ]; intros f y.
    intros [].
    destruct (Yeq_dec y (f x (or_introl eq_refl))) as [ E | C ].
    exists x, (or_introl eq_refl); auto.
    intros H.
    assert (In y (list_In_map l (fun x Hx => f x (or_intror Hx)))) as Hy.
      destruct H; auto.
      contradict C; auto.
    destruct IH with (f := fun x Hx => f x (or_intror Hx))
                     (1 := Hy)
      as (u & Hu & ?); subst.
    exists u, (or_intror Hu); auto.
  Qed.

End list_In_map.

Fact map_list_In_map X Y Z (g : Y -> Z) (l : list X) f : 
   map g (list_In_map l f) = list_In_map l (fun x Hx => g (f x Hx)).
Proof.
  revert f; induction l; simpl; intros f; f_equal; auto.
Qed.

(* maximum of a list *)

Definition lmax := fold_right max 0.

Fact lmax_prop n l : In n l -> n <= lmax l.
Proof.
  induction l as [ | x l IHl ].
  intros [].
  intros [ [] | ]; simpl.
  apply le_max_l.
  intros H; apply le_trans with (1 := IHl H).
  apply le_max_r.
Qed.

Fact lmax_inv l : { l = nil } + { In (lmax l) l }.
Proof.
  induction l as [ | x l IHl ].
  left; auto.
  right.
  simpl.
  destruct IHl.
  subst; left; simpl.
  rewrite max_0_r; auto.
  destruct (le_lt_dec x (lmax l)).
  right; rewrite max_r; auto.
  left; rewrite max_l; auto.
  omega.
Qed.
  
Fact fall_list_bound X (R : X -> nat -> Prop) l : 
      (forall x, In x l -> exists n, R x n)
   ->  exists m, forall x, In x l -> exists n, n <= m /\ R x n.
Proof.
  intros H. 
  induction l as [ | x l IHl ].
  exists 0; intros _ [].
  destruct IHl as (n & Hn).
  intros; apply H; right; auto.
  destruct (H _ (or_introl eq_refl)) as (m & Hm).
  exists (max m n).
  intros y [ H1 | H1 ].
  subst y; exists m; split; auto; apply le_max_l.
  destruct (Hn _ H1) as (k & H2 & H3).
  exists k; split; auto.
  apply le_trans with (1 := H2), le_max_r.
Qed.

Fact In_map_nat_find X (f : X -> nat) x l : In x (map f l) -> { u | x = f u /\ In u l }.
Proof.
  induction l as [ | y l IH ].
  intros [].
  destruct (eq_nat_dec x (f y)) as [ | C ].
  exists y; split; auto; left; auto.
  intros H.
  destruct IH as (u & H1 & H2).
  destruct H; auto; contradict C; auto.
  exists u; split; auto; right; auto.
Qed.

Section list_product.

  Variable (X Y Z : Type) (c : X -> Y -> Z).

  Definition list_product (l : list X) (m : list Y) 
    := flat_map (fun x => map (c x) m) l.

  Fact in_list_product l m p : In p (list_product l m) <-> exists x y, p = c x y /\ In x l /\ In y m.
  Proof.
    unfold list_product.
    rewrite in_flat_map.
    split.
    intros (x & H1 & H2).
    apply in_map_iff in H2.
    destruct H2 as (y & H2 & H3).
    exists x, y; auto.
    intros (x & y & ? & ? & ?); subst.  
    exists x; split; auto.
    apply in_map_iff; exists y; auto.
Qed.

End list_product.

Section list_eq_dec.

  Variable (X : Type).
  
  Implicit Types l : list X.
  
  Definition list_unary_dec l P : (forall x, In x l -> { P x } + { ~ P x }) -> { x | In x l /\ ~ P x } + { forall x, In x l -> P x }.
  Proof.
    induction l as [ | x l IH ].
    right; intros _ [].
    intros H.
    destruct IH as [ (y & H1 & H2) | IH ].
    intros; apply H; right; auto.
    left; exists y; split; auto; right; auto.
    destruct (H _ (or_introl eq_refl)).
    right; intros ? [ [] | ]; auto.
    left; exists x;split; auto; left; auto.
  Qed.
  
  Definition list_unary_dec_inv l P : (forall x, In x l -> { P x } + { ~ P x }) -> { x | In x l /\ P x } + { forall x, In x l -> ~ P x }.
  Proof.
    intros HP.
    destruct (list_unary_dec l (fun x => ~ P x)) as [ (x & H1 & H2) | H ]; simpl.
    intros x Hx; destruct (HP x Hx); tauto.
    left; exists x; split; auto; destruct (HP _ H1); tauto.
    right; auto.
  Qed.

  Definition list_eq_dec l m : (forall x y, In x l -> In y m -> { x = y } + { x <> y }) -> { l = m } + { l <> m }.
  Proof.
    revert m; induction l as [ | x l IH ]; intros [ | y m ].
    left; auto.
    right; discriminate.
    right; discriminate.
    intros H;
    destruct (H x y) as [ ? | C ]; try (simpl; tauto).
    subst.
    destruct (IH m) as [ ? | C ].
    intros; apply H; right; auto.
    subst; left; auto.
    right; intros H1; injc H1; apply C; auto.
    right; intros H1; injc H1; apply C; auto.
  Qed.
  
End list_eq_dec.

Section fall_In.
  
  Variable X : Type.

  Implicit Type (l : list X).

  Fixpoint fall_In l (P : forall x, In x l -> Prop) { struct l } : Prop.
  Proof.
    destruct l as [ | k l ].
    exact True.
    apply and.
    apply (P _ (or_introl eq_refl)).
    apply (fall_In l).
    intros x ?; apply (P x); right; assumption.
  Defined.

  Fact fall_In_eq l P : @fall_In l P <-> forall x Hx, P x Hx.
  Proof.
    split; revert P.

    induction l as [ | x l IH ]; simpl; intros P HP.
    destruct Hx.
    destruct HP.
    intros ? [ [] | y ]; auto.
    apply (IH (fun x Hx => P _ (or_intror Hx))); auto.
  
    induction l as [ | x l IH ]; simpl; intros P HP; auto.
  Qed.

  Fact fall_In_ext l R S : (forall x (Hx : In x l), R x Hx = S x Hx) -> @fall_In l R = @fall_In l S.
  Proof.
    revert R S; induction l as [ | x l IHl ]; simpl; intros R S HRS; auto.
    f_equal. 
    apply HRS.
    apply IHl; intros; apply HRS.
  Qed.
 
End fall_In.

