(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements 

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel
 
   and Lambda Calculus, Types and Models by JL Krivine

*)

Require Import List Relations Arith Max Omega.

Require Import utils pos vec.
Require Import uterm uterm_eta uterm_beta uterm_beta_eta uterm_beta_eq uterm_normal.
Require Import itypes.

Set Implicit Arguments.

Section system_D_omega.

  Variable X : Set.

  Notation stype := (stype X).
  Notation itype := (list stype).

  (* This is system DOmega, it is used to characterize 
       1/ head normalization
       2/ leftmost/beta normalization
       3/ beta-eta normalization
   *)

  (* Again, because of rule in_ta_int, at some points, we need to reason 
     on the height of proofs so we record it in the inductive predicate *)

  Reserved Notation " e '|-' x '#' n '#' t " (at level 70).

  Inductive it_assign_n : nat -> forall n, vec itype n -> uterm n -> itype -> Prop := 

    | in_ta_ax  : forall h n e p t, @vec_pos _ n e p <it t::nil ->     e |- £ p #S h# (t::nil)
    
    | in_ta_int : forall h n (e : vec _ n) u t,  ( forall x, In x t -> e |- u   #h# x::nil )
                                        ->                             e |- u   #S h# t
                                        
    | in_ta_abs : forall h n (e : vec _ n) s u t,                   s##e |- u    #h# t::nil 
                                        ->                             e |- \\ u #S h# s-->t::nil
                                        
    | in_ta_app : forall h n (e : vec _ n) f a s t,                    e |- f    #h# s-->t::nil
                                        ->                             e |- a    #h# s
                                        ->                             e |- f@a  #S h# t::nil
                                        
  where " e |- x # h # t " := (@it_assign_n h _ e x t).
  
  Fact it_assign_n_le a b : a <= b -> forall n e (f : uterm n) s, e |- f #a# s -> e |- f #b# s.
  Proof.
    induction 1 as [ | b H IH ]; auto; intros n e f s H1.
    specialize (IH _ _ _ _ H1); revert IH; clear H1 H a.
    induction 1 as [ | | | ? ? ? ? ? s ].
    constructor 1; auto.
    constructor 2; auto.
    constructor 3; auto.
    constructor 4 with s; auto.
  Qed.

  Definition it_assign n e (f : uterm n) t := exists h, e |- f #h# t.

  Notation " e '|-' x ':t' t " := (@it_assign _ e x t) (at level 70).

  Fact ita_ax n e p t : @vec_pos _ n e p <it t::nil -> e |- £ p :t (t::nil).
  Proof.  exists 1; constructor 1; auto. Qed.

  Fact ita_int n (e : vec _ n) u t :  ( forall x, In x t -> e |- u :t x::nil )
                            ->                              e |- u :t t.
  Proof.
    intros H.
    apply fall_list_bound in H.
    destruct H as (m & H).
    exists (S m).
    constructor 2.
    intros x Hx.
    destruct (H x Hx) as (q & H1 & H2).
    revert H2; apply it_assign_n_le; auto.
  Qed.

  Fact ita_abs n (e : vec _ n) s u t :   s##e |- u    :t t::nil 
                                      ->    e |- \\ u :t s-->t::nil.
  Proof.
    intros (h & ?); exists (S h); constructor 3; auto.
  Qed.

  Fact ita_app n (e : vec _ n) f a s t :    e |- f      :t s-->t::nil
                                      ->    e |- a      :t s
                                      ->    e |- f @ a  :t t::nil.
  Proof.
    intros (h1 & H1) (h2 & H2).
    exists (S (max h1 h2)).
    constructor 4 with s.
    revert H1; apply it_assign_n_le, le_max_l.
    revert H2; apply it_assign_n_le, le_max_r.
  Qed.

  Section it_assign_ind.
    
    Variable P : forall n, vec itype n -> uterm n -> itype -> Prop.

    Notation " e '|=' f ':t' t " := (@P _ e f t) (at level 70).

    Hypothesis HS1 : forall n e (p : pos n) t,
                                    vec_pos e p <it t::nil
                                 -> e |= £ p :t t::nil.

    Hypothesis HS2 : forall n e (f : uterm n) t,
                                    (forall u, In u t -> e |- f :t u::nil)
                                 -> (forall u, In u t -> e |= f :t u::nil)
                                 ->                      e |= f :t t.

    Hypothesis HS3 : forall n e (f : uterm (S n)) l t,
                                    l ## e |- f   :t t::nil
                                 -> l ## e |= f   :t t::nil
                                 ->      e |= \\f :t l-->t::nil.

    Hypothesis HS4 : forall n e (f a : uterm n) l t,
                                    e |- f     :t l-->t::nil
                                 -> e |= f     :t l-->t::nil
                                 -> e |- a     :t l
                                 -> e |= a     :t l
                                 -> e |= f @ a :t t::nil.

    Theorem it_assign_ind n e (f : uterm n) t : e |- f :t t -> e |= f :t t.
    Proof.
      intros (h & Hh).
      induction Hh.

      apply HS1; auto.
      apply HS2; auto; exists h; auto.
      apply HS3; auto; exists h; auto.
      apply HS4 with s; auto; exists h; auto.
    Qed.
 
  End it_assign_ind.
  
  Section Lemma_1_5.
  
    Let it_assign_n_inv_0 n e (f : uterm n) s : ~ e |- f #0# s.
    Proof. inversion 1. Qed.

    Let it_assign_n_inv_S h n e (f : uterm n) s : 
       e |- f #S h# s -> (exists p t, f = £ p /\ s = t::nil /\ vec_pos e p <it s)
                      \/ (forall x, In x s -> e |- f #h# x::nil)
                      \/ (exists g l t, f = \\ g /\ s = l-->t::nil /\ l##e |- g #h# t::nil)
                      \/ (exists g a l t, f = g @ a /\ s = t::nil /\ e |- a #h# l /\ e |- g #h# l-->t::nil). 
    Proof.
      gen_eq (S h); intros i H1 H2.
      generalize i n e f s H2 h H1.
      clear n h e f s i H1 H2.
      induction 1 as [ h n e p t | h n e f t | h n e s u t | h n e f a s t ]; intros i E; injc E.
      left; exists p, t; auto.
      right; left; auto.
      do 2 right; left; exists u, s, t; auto.
      do 3 right; exists f, a, s, t; auto.
    Qed.
    
    Lemma ita_var_inv n e (x : pos n) s : e |- £ x :t s <-> vec_pos e x <it s.
    Proof.
      split.
      
      intros (h & H).
      revert n e x s H.
      induction h as [ | h IHh ]; intros n e x s H.

      apply it_assign_n_inv_0 in H; destruct H.
      
      apply it_assign_n_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3) 
                  | [ G1
                  | [ (f & l & t & G1 & G2 & G3)
                    | (f & a & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      apply ut_var_inj in G1; subst; auto.
      apply le_it_int; auto.
      discriminate G1.
      discriminate G1.

      intros H.
      exists 2.
      constructor 2.
      constructor 1. 
      apply le_it_trans with (1 := H).
      apply le_it_ax; auto.
    Qed.
      
    Lemma ita_app_inv n e (a b : uterm n) s : e |- a @ b :t s::nil <-> exists l, e |- a :t l-->s::nil /\ e |- b :t l.
    Proof.
      split.
      
      intros (h & H).
      revert h n e a b s H.
      induction h as [ | h IHh ]; intros n e a b s H.

      apply it_assign_n_inv_0 in H; destruct H.
    
      apply it_assign_n_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3) 
                  | [ G1
                  | [ (f & l & t & G1 & G2 & G3)
                    | (f & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      discriminate G1.
      apply IHh, G1; left; auto.
      discriminate G1.
      injc G1; injc G2.
      exists l; split; exists h; auto.
      
      intros (l & (h1 & H1) & h2 & H2).
      exists (S (max h1 h2)). 
      constructor 4 with l.
      revert H1; apply it_assign_n_le, le_max_l.
      revert H2; apply it_assign_n_le, le_max_r.
    Qed.
    
    Lemma typable_app_rev_inv n e (f : uterm n) l t : e |- f *@ l :t t::nil -> exists t, e |- f :t t::nil.
    Proof.
      revert t; induction l as [ | x l IHl ]; simpl.
      intros t; exists t; auto.
      intros t Ht.
      apply ita_app_inv in Ht.
      destruct Ht as (m & H1 & H2).
      apply IHl with (1 := H1). 
    Qed.
    
    Lemma ita_abs_inv n e (f : uterm (S n)) s : e |- \\ f :t s::nil <-> exists l t, s = l-->t /\ l##e |- f :t t::nil.
    Proof.
      split.
      
      intros (h & H).
      revert h n e f s H.
      induction h as [ | h IHh ]; intros n e f s H.
      
      apply it_assign_n_inv_0 in H; destruct H.
      
      apply it_assign_n_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3) 
                  | [ G1
                  | [ (g & l & t & G1 & G2 & G3)
                    | (g & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      discriminate G1.
      apply IHh, G1; left; auto.
      injc G1; injc G2; exists l, t; split; auto; exists h; auto.
      discriminate G1.
      
      intros (l & t & ? & h & H); subst.
      exists (S h); constructor 3; auto.
    Qed.
    
    Lemma ita_int_inv n e (f : uterm n) s : e |- f :t s <-> forall t, In t s -> e |- f :t t::nil.
    Proof.
      split.
      
      intros (h & H).
      revert h n e f s H.
      induction h as [ | h IHh ]; intros n e f s H.
      
      apply it_assign_n_inv_0 in H; destruct H.
      
      apply it_assign_n_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3) 
                  | [ G1
                  | [ (g & l & t & G1 & G2 & G3)
                    | (g & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      intros u [ ? | [] ]; subst u; exists 1; constructor 1; auto.
      exists h; auto.
      intros u [ ? | [] ]; subst u; exists (S h); constructor 3; auto.
      intros v [ ? | [] ]; subst v; exists (S h); constructor 4 with l; auto.
      
      intros H.
      apply fall_list_bound in H.
      destruct H as (m & Hm).
      exists (S m).
      constructor 2.
      intros x Hx; destruct (Hm x Hx) as (g & ? & G).
      revert G; apply it_assign_n_le; auto.
    Qed.

  End Lemma_1_5.    
    
    (** This lemma proves several things at once
     **    1/ only the value of the environment on the variable occuring in f matter
     **    2/ strengthening the hyps or weakening the concs preserve typing judgements  
     **)
    
  Lemma ita_weak n e1 e2 (f : uterm n) s t : 
          (forall p, 0 < ut_occ f p -> vec_pos e2 p <it vec_pos e1 p)
        -> s <it t 
        -> e1 |- f :t s 
        -> e2 |- f :t t.
  Proof.
    intros H1 H2 Hh.
    revert Hh t e2 H2 H1.
    induction 1 as [ n e1 p t 
                   | n e1 f t H1 IH1
                   | n e1 s u t H1 IH1 
                   | n e1 f a s t H1 IH1 H2 IH2 H3 ]
      using it_assign_ind; intros w e2 Hw He2.

    rewrite le_it_int_eq in Hw.
    apply ita_int.
    intros y Hy.
    apply ita_ax.
    apply le_it_trans with (2 := Hw _ Hy),
          le_it_trans with (2 := H). 
    apply He2; simpl.
    destruct (pos_eq_dec p p) as [ | [] ]; auto.

    rewrite le_it_any_inv in Hw.
    apply ita_int.
    intros y Hy; destruct (Hw _ Hy) as (x & Hx1 & Hx2).
    apply (IH1 _ Hx1); auto.

    rewrite le_it_int_eq in Hw.
    apply ita_int.
    intros x Hx; specialize (Hw _ Hx).
    rewrite le_it_arr_inv in Hw.
    destruct Hw as (s' & t' & ? & Hw1 & Hw2); subst.
    apply ita_abs.
    apply IH1; auto.
    intros p Hp; pos_inv p; simpl; auto.

    rewrite le_it_int_eq in Hw.
    simpl in He2.
    apply ita_int.
    intros x Hx; specialize (Hw _ Hx).
    apply ita_app with s.
    apply IH1.
    apply le_it_arr; auto.
    apply le_it_refl.
    intros; apply He2; omega.
    apply IH2.
    apply le_it_refl.
    intros; apply He2; omega.
  Qed.

  Corollary ita_weak_left n e1 e2 (f : uterm n) t : 
          (forall p, vec_pos e2 p <it vec_pos e1 p)
        -> e1 |- f :t t 
        -> e2 |- f :t t.
  Proof.
    intro; apply ita_weak; auto; apply le_it_refl.
  Qed.

  Corollary ita_weak_right n e (f : uterm n) s t : 
           s <it t 
        -> e |- f :t s 
        -> e |- f :t t.
  Proof.
    intro; apply ita_weak; auto; intros; apply le_it_refl.
  Qed.
  
  Fact ita_weak_sg n e (f : uterm n) x t : In x t -> e |- f :t t -> e |- f :t x::nil.
  Proof.
    intro; apply ita_weak_right, le_it_ax; auto.
  Qed.

  Fact ita_stype_itype n1 e1 (f1 : uterm n1) n2 e2 (f2 : uterm n2) t :
       (forall x, In x t -> e1 |- f1 :t x::nil -> e2 |- f2 :t x::nil)
    -> (                    e1 |- f1 :t t      -> e2 |- f2 :t t     ).
  Proof.
    intros H1 H.
    apply ita_int.
    intros x Hx.
    apply ita_weak_sg with (1 := Hx) in H; auto.
  Qed.

  Notation " x '~[' R ']~' y " := (@ut_ren_rel _ _ R x y).

  Fact ita_ren_rel n m (R : ren_rel n m) e1 e2 f1 f2 t :
        f1 ~[R]~ f2
     -> (forall p q, R p q -> vec_pos e1 p = vec_pos e2 q)
     -> e1 |- f1 :t t -> e2 |- f2 :t t.
  Proof.
    intros H H1.
    apply ita_stype_itype.
    intros x _.
    revert H e1 e2 x H1.
    clear t.

    induction 1 as [ n m R f g H1 IH1 
                   | n m R f g a b H1 IH1 H2 IH2 
                   | n m R p q H1 ]; 
      intros e1 e2 t He H.

    apply ita_abs_inv in H.
    destruct H as (l & s & ? & H); subst.
    apply ita_abs.
    revert H; apply IH1.
    intros p q; pos_inv p; pos_inv q; try (destruct 1; fail); simpl; auto.

    apply ita_app_inv in H.
    destruct H as (l & G1 & G2).
    apply ita_app with l.
    revert G1; apply IH1; auto.
    apply ita_int.
    intros x Hx.
    apply ita_weak_sg with (1 := Hx) in G2. 
    revert G2; apply IH2; auto.

    apply ita_var_inv in H.
    apply ita_ax.
    rewrite <- (He _ _ H1); auto.
  Qed.

  Fact ita_lift l n e (f : uterm n) t : l##e |- ut_lift f :t t <-> e |- f :t t.
  Proof.
    split.

    apply ita_ren_rel with (R := fun p q => p = pos_nxt _ q).
    apply ut_ren_rel_sym, ut_ren_rel_eq_ren.
    intros p q; pos_inv p; simpl.
    discriminate 1.
    intros H; apply pos_nxt_inj in H.
    subst; auto.

    apply ita_ren_rel with (R := fun p q => q = pos_nxt _ p).
    apply ut_ren_rel_eq_ren.
    intros p q; pos_inv q; simpl.
    discriminate 1.
    intros H; apply pos_nxt_inj in H.
    subst; auto.
  Qed.

  Fact ita_subst_rel n1 n2 (R : ren_rel n1 n2) m1 m2 (S : ren_rel m1 m2) e1 e2 f1 f2 t s1 s2 :
        f1 ~[R]~ f2
     -> (forall p q, S p q -> vec_pos e1 p = vec_pos e2 q)
     -> (forall p q, R p q -> s1 p ~[S]~ s2 q)
     -> e1 |- ut_subst f1 s1 :t t -> e2 |- ut_subst f2 s2 :t t.
  Proof.
    intros H1 H2 H3.
    apply ita_stype_itype.
    intros x _.
    revert H1 m1 m2 S e1 e2 x s1 s2 H2 H3.
    clear t.

    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ]; 
      intros m1 m2 S e1 e2 t s1 s2 H3 H4.

    simpl; rewrite ita_abs_inv.
    intros (l & x & ? & H5); subst.
    apply ita_abs; revert H5.
    apply (IH1 _ _ (ren_rel_lift S)).
    intros p q; pos_inv p; pos_inv q; try destruct 1; simpl; auto.
    intros p q; pos_inv p; pos_inv q; try destruct 1.
    unfold uts_lift; simpl; constructor 3; reflexivity.
    unfold ren_rel_lift at 1; simpl; intros H.
    do 2 rewrite uts_lift_tail.
    unfold ut_lift.
    apply ut_ren_rel_ren with S; auto.
   
    simpl; rewrite ita_app_inv.
    intros (l & H5 & H6).
    apply ita_app with l.
    revert H5.
    apply (IH1 _ _ S); auto.
    revert H6.
    apply ita_stype_itype.
    intros ? _; apply (IH2 _ _ S); auto.

    simpl.
    apply ita_ren_rel with S; auto.
  Qed.

  Definition ienv_subst n (s : X -> _) e := vec_set_pos (fun p => itype_subst s (@vec_pos _ n e p)).

  Arguments ienv_subst {n} s e /.

  Lemma ita_itype_subst n e (f : uterm n) t s : 
       e |- f :t t -> ienv_subst s e |- f :t itype_subst s t.
  Proof.
    induction 1 as [ n e x t H 
                   | n e u t H1 IH1 
                   | n e l f c H1 IH1
                   | n e f a l c H1 IH1 H2 IH2 ]
      using it_assign_ind; simpl.

    apply ita_ax; rewrite vec_pos_set.
    apply (le_it_itype_subst s H).

    apply ita_int.
    intros x; unfold itype_subst; rewrite in_map_iff.
    intros (q & ? & Hq); subst.
    apply (IH1 _ Hq).

    rewrite stype_subst_arr.
    apply ita_abs, IH1.
    
    apply ita_app with (itype_subst s l).
    unfold itype_subst in IH1; simpl in IH1.
    rewrite stype_subst_arr in IH1.
    apply IH1.
    apply IH2.
  Qed.
  
  Lemma ita_pre_eta n e (f : uterm n) l t : l##e |- ut_lift f @ £ pos0 :t t::nil
                                         ->    e |- f :t l-->t::nil.
  Proof.
    intros H.
    apply ita_app_inv in H.
    destruct H as (m & H1 & H2).
    apply ita_var_inv in H2; simpl in H2.
    apply ita_lift in H1.
    revert H1.
    apply ita_weak_right.
    apply le_it_arr; auto.
    apply le_it_refl.
  Qed.

  (* eta redex reduction preserves typing judgments *)
  
  Lemma ita_eta_redex n e (f : uterm n) t : e |- \\ (ut_lift f @ £ pos0) :t t
                                         -> e |- f :t t.
  Proof.
    apply ita_stype_itype; revert t; intros _ t _ H.
    apply ita_abs_inv in H.
    destruct H as (l & s & ? & H); subst.
    revert H; apply ita_pre_eta.
  Qed.

  (* Typing judgments are preserved by eta reduction *)

  Theorem ita_eta n e (f g : uterm n) t : f -e-> g -> e |- f :t t -> e |- g :t t.
  Proof.
    intros H. 
    apply ita_stype_itype.
    intros x _; revert x; clear t. 
    induction H as [ n f 
                   | n f g a H1 IH1 
                   | n a b f H1 IH1 
                   | n f g H1 IH1 ]; intros t Ht.

    apply ita_eta_redex; auto.

    rewrite ita_app_inv in Ht.
    destruct Ht as (l & H3 & H4).
    apply ita_app with l; auto.

    rewrite ita_app_inv in Ht.
    destruct Ht as (l & H3 & H4).
    apply ita_app with l; auto.
    revert H4; apply ita_stype_itype.
    intros; auto.

    rewrite ita_abs_inv in Ht.
    destruct Ht as (l & c & ? & Ht); subst.
    apply ita_abs; auto.
  Qed.

  Theorem ita_eta_rt n e (f g : uterm n) t : f -e>> g -> e |- f :t t -> e |- g :t t.
  Proof.
    induction 1; auto.
    apply ita_eta; auto.
  Qed.

  (* but eta redex expansion preserve only typability *)

  Lemma ita_eta_redex_rev n e 
           (f : uterm n) t (l : X -> itype) (c : X -> stype) :
       e |- f :t t::nil  
    -> ienv_subst (fun x => l x --> c x) e |- \\ (ut_lift f @ £ pos0) :t stype_subst (fun x => l x --> c x) t :: nil.
  Proof.
    intros Ht.
    destruct t as [ x | lt ct ].

    apply ita_itype_subst with (s := fun x => l x --> c x) in Ht.
    rewrite stype_subst_var.
    apply ita_abs, ita_app with (l x).
    apply ita_lift.
    simpl in Ht.
    rewrite stype_subst_var in Ht.
    apply Ht.
    apply ita_int.
    intros.
    apply ita_ax; simpl.
    apply le_it_ax; auto.

    rewrite stype_subst_arr.
    apply ita_abs, ita_app with (itype_subst (fun x => l x --> c x) lt).
    apply ita_lift.
    apply ita_itype_subst with (s := fun x => l x --> c x) in Ht.
    simpl in Ht; rewrite stype_subst_arr in Ht.
    apply Ht.
    apply ita_int.
    intros.
    apply ita_ax; simpl.
    apply le_it_ax; auto.
  Qed.

  Lemma ita_eta_rev n e (f g : uterm n) t l c :
       f -e-> g -> e |- g :t t 
    -> ienv_subst (fun x => l x --> c x) e |- f :t itype_subst (fun x => l x --> c x) t.
  Proof.

    (* first replace itypes with stypes*)

    intros H He.
    apply ita_int.
    simpl.
    intros x; rewrite in_map_iff.
    intros (u & ? & Hu); subst.
    rewrite ita_int_inv in He.
    specialize (He _ Hu).
    clear t Hu.
    change (ienv_subst (fun x => l x --> c x) e |- f :t stype_subst (fun x : X => l x --> c x) u :: nil).
    revert u He.

    (* Now we proceed with stypes *)

    intros t; revert H e t.
    induction 1 as [ n f Hf | n f g a H IH | n a b f H IH | n f g H IH ]; 
      intros e t He.
 
    apply ita_eta_redex_rev, He.

    rewrite ita_app_inv in He.
    destruct He as (m & Hl & Ha).
    apply IH in Hl.
    rewrite stype_subst_arr in Hl.
    apply ita_itype_subst with (s := fun x => l x --> c x) in Ha. 
    apply ita_app with (1 := Hl); auto.

    rewrite ita_app_inv in He.
    destruct He as (m & Hl & Ha).
    apply ita_itype_subst with (s := fun x => l x --> c x) in Hl. 
    simpl in Hl; rewrite stype_subst_arr in Hl.
    apply ita_app with (1 := Hl).
    apply ita_int; intros x; simpl.
    rewrite in_map_iff.
    intros (u & ? & Hu); subst.
    rewrite ita_int_inv in Ha.
    apply IH; auto.

    apply ita_abs_inv in He.
    destruct He as (m & t' & ? & He); subst.
    rewrite stype_subst_arr.
    apply ita_abs, (IH _ _ He).
  Qed.

  Definition typable_polarity_free n (f : uterm n) :=
     exists e t, e |- f :t t 
              /\ itype_w_free_pos t 
              /\forall p, itype_w_free_neg (vec_pos e p).

  Definition typable_wo_omega n (f : uterm n) :=
     exists e t, e |- f :t t 
              /\ itype_w_free t 
              /\forall p, itype_w_free (vec_pos e p).

  Fact typable_wo_omega_typable_polarity_free n (f : uterm n) :
      typable_wo_omega f -> typable_polarity_free f. 
  Proof.
    intros (e & t & H1 & H2 & H3).
    exists e, t; split; auto; split.
    split.
    apply H2.
    intros u Hu; apply (proj2 H2); auto.
    intros p u Hu; apply (proj2 (H3 p)); auto.
  Qed.

  Corollary typeable_pf_eta_rt n (f g : uterm n) :
     f -e>> g -> typable_polarity_free f -> typable_polarity_free g.
  Proof.
    intros H1 (e & t & H2 & H3 & H4).
    exists e, t; split; auto.
    revert H1 H2; apply ita_eta_rt.
  Qed.

  (* eta expansion does not preserve typing judgments BUT

     eta expansion preserves polarity free typability 
     which characterize (leftmost) normalizable terms *)

  Theorem typable_pf_eta_rev n (f g : uterm n) :
     f -e-> g -> typable_polarity_free g -> typable_polarity_free f.
  Proof.
    intros H (e & t & H1 & H2 & H3).
    generalize (itype_w_free_pos_to_var H2); intros x.
    generalize (ita_eta_rev (fun _ => µ x::nil) (fun _ => µ x) H H1).
    match goal with |- ?x |- _ :t ?r -> _ => exists x, r end.
    split; auto.
    split.
    apply itype_w_free_pos_subst; auto.
    intros p; unfold ienv_subst; rewrite vec_pos_set.
    apply itype_w_free_neg_subst; auto.
  Qed.

  Theorem typable_pf_eta_rt_rev n (f g : uterm n) :
     f -e>> g -> typable_polarity_free g -> typable_polarity_free f.
  Proof.
    induction 1; auto.
    apply typable_pf_eta_rev; auto.
  Qed.

  (* So we are done with eta reduction/expansion, it preserves
     polarity free typability and hence allows to show that
     beta-eta normalizable is the same as beta normalizable *)

  (* Now let us proceed we beta reduction *)

  Lemma ita_subst_1 n (f : uterm n) m s e t : 
          e |- @ut_subst n f m s :t t::nil 
     -> exists e', e' |- f :t t::nil 
                 /\ forall p, e |- s p :t vec_pos e' p.
  Proof.
    revert m s e t.
    induction f as [ n f | n a IHa b IHb | n v ]; intros m s e t; simpl.

    (* case of a \\ *)

    rewrite ita_abs_inv.
    intros (l & r & ? & H1); subst. 
    apply IHf in H1.
    destruct H1 as (e' & H1 & H2); revert H1 H2.
    rewrite (vec_head_tail e').
    generalize (vec_head e') (vec_tail e'); clear e'.
    intros l' e' H1 H2.     
    exists e'; split.
    specialize (H2 pos0).
    unfold uts_lift in H2; simpl in H2.
    apply ita_abs; auto.
    revert H1; apply ita_weak.  
    intros p Hp; pos_inv p; simpl; auto.
    apply ita_var_inv in H2; simpl in H2; auto.
    apply le_it_refl.
    apply le_it_refl.
    intros p; specialize (H2 (pos_nxt _ p)).
    unfold uts_lift in H2; simpl in H2.
    revert H2; apply ita_lift.

    (* case of a @ b *)
 
    rewrite ita_app_inv.
    intros (l & Ha & Hb).
    apply IHa in Ha.
    destruct Ha as (ea & Ha1 & Ha2).
    rewrite ita_int_inv in Hb.
    generalize (fun x Hx => IHb _ _ _ _ (Hb x Hx)); intros IHb'.
    apply exists_Forall2 in IHb'.
    destruct IHb' as (leb & Hleb).

    (* this is the intersection of all environments *)
    set (e' := vec_set_pos (fun p => flat_map (fun e => vec_pos e p) leb)).
    assert (e' |- b :t l /\ forall p, e |- s p :t vec_pos e' p ) as He'.
      unfold e'; split.
      apply ita_int.
      intros x Hx.
      destruct (Forall2_left_inv Hleb _ Hx) as (u & H1 & H2 & H3).
      revert H2; apply ita_weak_left.
      intros p; rewrite vec_pos_set.
      apply le_it_any_inv.
      intros z Hz; exists z; split.
      apply in_flat_map; exists u; split; auto.
      apply le_it_refl.
      intros p; rewrite vec_pos_set.
      apply ita_int.
      intros x; rewrite in_flat_map.
      intros (u & H1 & H2).
      destruct (Forall2_right_inv Hleb _ H1) as (y & H3 & H4 & H5).
      generalize (H5 p).
      apply ita_weak_right.
      apply le_it_ax; auto.
    generalize e' He'.
    clear leb Hleb e' He' Hb.
    intros eb (Hb1 & Hb2).

    exists (vec_set_pos (fun p => vec_pos ea p ++ vec_pos eb p)); split.

    apply ita_app with l.

    revert Ha1; apply ita_weak_left.
    intros; rewrite vec_pos_set; simpl; apply le_it_app_left.
    revert Hb1; apply ita_weak_left.
    intros; rewrite vec_pos_set; simpl; apply le_it_app_right.

    intros p; rewrite vec_pos_set.
    apply ita_int.
    intros x Hx; apply in_app_or in Hx.
    destruct Hx; [ generalize (Ha2 p) |  generalize (Hb2 p) ];
      apply ita_weak_right, le_it_ax; auto.

    (* case of a variable *)

    exists (vec_set_pos (fun p => if pos_eq_dec v p then t::nil else nil)).
    split. 
    rewrite ita_var_inv, vec_pos_set.
    destruct (pos_eq_dec v v) as [ _ | [] ]; auto.
    apply le_it_refl.
    intros p; rewrite vec_pos_set.
    destruct (pos_eq_dec v p); subst; auto.
    apply ita_int; destruct 1.
  Qed.
(*
  Let gt0_choose x y : 0 < x + y -> 0 < x \/ 0 < y.
  Proof. omega. Qed.
*)

  Lemma ita_subst_2 n (f : uterm n) m s e t e' : 
          e |- f :t t::nil
      -> (forall p, e' |- s p :t vec_pos e p)
      ->  e' |- @ut_subst n f m s :t t::nil.
  Proof.
    revert m s e t e'.
    induction f as [ n f | n a IHa b IHb | n v ]; intros m s e t e'; simpl.
 
    rewrite ita_abs_inv.
    intros (l & t' & ? & H1) H2; subst.
    apply ita_abs, IHf with (1 := H1).
    intros p; pos_inv p; unfold uts_lift; simpl.
    apply ita_int; intros x Hx; apply ita_ax, le_it_ax, Hx.
    apply ita_lift, H2.

    rewrite ita_app_inv.
    intros (l & H1 & H2) H3.
    apply ita_app with l.
    apply IHa with (1 := H1); auto.
    revert H2; apply ita_stype_itype.
    intros x _ H2; apply IHb with (1 := H2); auto.

    rewrite ita_var_inv.
    intros H1 H2.
    generalize (H2 v).
    apply ita_weak_right; auto.
  Qed.

  Theorem ita_subst n (f : uterm n) m s e t : 
          e |- @ut_subst n f m s :t t 
      <-> exists e', e' |- f :t t 
         /\ forall p, e |- s p :t vec_pos e' p.
  Proof.
    split.

    intros H.
    rewrite ita_int_inv in H.
    generalize (fun x Hx => ita_subst_1 _ _ (H x Hx)); clear H; intros H.
    apply exists_Forall2 in H.
    destruct H as (l & Hl).
    exists (vec_set_pos (fun p => flat_map (fun e => vec_pos e p) l)).
    split.
    apply ita_int.
    intros x Hx.
    destruct (Forall2_left_inv Hl _ Hx) as (u & H1 & H2 & H3).
    revert H2; apply ita_weak_left.
    intros p; rewrite vec_pos_set.
    apply le_it_any_inv.
    intros z Hz; exists z; split.
    apply in_flat_map; exists u; split; auto.
    apply le_it_refl.
    intros p; rewrite vec_pos_set.
    apply ita_int.
    intros x; rewrite in_flat_map.
    intros (u & H1 & H2).
    destruct (Forall2_right_inv Hl _ H1) as (y & H3 & H4 & H5).
    generalize (H5 p).   
    apply ita_weak_right, le_it_ax; auto.

    intros (e' & H1 & H2).
    apply ita_int.
    intros x Hx.
    apply ita_subst_2 with e'; auto.
    revert H1; apply ita_weak_right, le_it_ax; auto.
  Qed.

  Corollary ita_reduce n (f : uterm (S n)) (a : uterm n) e t :
      e |- f[a] :t t <-> exists l, l##e |- f :t t /\ e |- a :t l.
  Proof.
    unfold ut_reduce.
    rewrite ita_subst; split.

    intros (e' & H1 & H2); revert H1 H2.
    rewrite (vec_head_tail e').
    generalize (vec_head e') (vec_tail e'); clear e'.
    intros l e' H1 H2.
    assert (forall p, vec_pos e p <it vec_pos e' p) as H3.
      intros p.
      specialize (H2 (pos_nxt _ p)).
      unfold uts_first in H2; simpl in H2.
      apply ita_var_inv in H2; auto.
    exists l; split.
    revert H1; apply ita_weak_left.
    intros p; pos_inv p; simpl; auto.
    apply le_it_refl.
    apply (H2 pos0).

    intros (l & H1 & H2).
    exists (l##e); split; auto.
    intros p; pos_inv p; unfold uts_first; simpl; auto.
    apply ita_var_inv, le_it_refl.
  Qed.
  
  Theorem ita_subject_reduction n (f g : uterm n) e t : f ~b g -> e |- f :t t::nil <-> e |- g :t t::nil.
  Proof.
    intros H; revert H e t.
    induction 1 as [ f g H | | f g H IH | f g h H1 IH1 H2 IH2 ].
   
    (* this is for -b-> *)
   
    induction H as [ n f a | n a b g H1 IH1 | n a b g H1 IH1 | n f g H1 IH1 ]; intros e t.
 
    rewrite ita_app_inv.
    split.
    intros (l & H1 & H2).
    apply ita_abs_inv in H1.
    destruct H1 as (l' & t' & E & H1); injc E.
    apply ita_reduce; exists l'; auto.
    rewrite ita_reduce; intros (l & H1 & H2).
    exists l; split; auto; apply ita_abs; auto.

    do 2 rewrite ita_app_inv.
    split; intros (l & H3 & H4); exists l; 
      split; auto; revert H3; apply IH1.

    do 2 rewrite ita_app_inv.
    split; intros (l & H3 & H4); exists l; split; auto;
      revert H4; apply ita_stype_itype; intros ? ?; apply IH1.

    do 2 rewrite ita_abs_inv.
    split; intros (l & s & E & H3); exists l, s; subst; 
      split; auto; revert H3; apply IH1.

    (* reflexivity *)

    tauto.

    (* symmetry *)

    intros ? ?; symmetry; apply IH.

    intros ? ?; rewrite IH1; apply IH2.
  Qed.

  Corollary typable_pf_beta_rev n (f g : uterm n) :
      f -b-> g -> typable_polarity_free g -> typable_polarity_free f.
  Proof.
    intros H (e & t & H1 & H2 & H3).
    exists e, t; split; auto.
    revert H1.
    apply ita_stype_itype.
    intros x Hx.
    apply H2 in Hx.
    apply ita_subject_reduction.
    apply ut_beta_rt_eq, ut_beta_beta_rt, H.
  Qed.

  Fact typable_pf_beta_eta_rev n (f g : uterm n) :
      f -be-> g -> typable_polarity_free g -> typable_polarity_free f.
  Proof.  
    intros [ H | H ]; revert H.
    apply typable_pf_beta_rev.
    apply typable_pf_eta_rev. 
  Qed.

  Theorem typable_pf_beta_eta_rt_rev n (f g : uterm n) :
      f -be>> g -> typable_polarity_free g -> typable_polarity_free f.
  Proof.
    induction 1; auto.
    apply typable_pf_beta_eta_rev; auto.
  Qed.

  Section typing_of_normal_terms.
  
    Variable x : X.

    Let w_free_typable n (f : uterm n) t := exists e, e |- f :t t::nil /\ forall p, itype_w_free (vec_pos e p).
    
    Let w_free_typable_app n (f a : uterm n) h c : w_free_typable f ((h::nil)-->c) -> w_free_typable a h -> w_free_typable (f@a) c.
    Proof.
      intros (ef & H1 & H2) (ea & H3 & H4).
      exists (vec_set_pos (fun p => vec_pos ea p ++ vec_pos ef p)); split.
      apply ita_app with (h::nil).
      revert H1; apply ita_weak_left.
      intros p; rewrite vec_pos_set.
      apply le_it_app_right.
      revert H3; apply ita_weak_left.
      intros p; rewrite vec_pos_set.
      apply le_it_app_left.
      intros p; rewrite vec_pos_set.
      apply itype_w_free_app; auto.
    Qed.
    
    Let w_free_typable_var n (v : pos n) t : stype_w_free t -> w_free_typable (£ v) t.
    Proof.
      intros Ht.
      exists (vec_set_pos (fun _ => t::nil)).
      split.
      apply ita_ax, le_it_ax.
      rewrite vec_pos_set; left; auto.
      intros p; rewrite vec_pos_set.
      apply itype_w_free_sg; auto.
    Qed.

    Let ut_normal_typable_w_free_rec n (f : uterm n) : 
        ut_normal f -> (                 exists t, stype_w_free t /\ w_free_typable f t )
                    /\ ( ut_neutral f -> forall t, stype_w_free t -> w_free_typable f t ).
    Proof.
      induction 1 as [ n f Hf IHf | n f a Hf1 Hf2 Ha IHf IHa | n v ] using ut_normal_rect; split. 

      destruct (proj1 IHf) as (t & H2 & e & H3 & H1).
      revert H1 H3; rewrite (vec_head_tail e).
      generalize (vec_head e) (vec_tail e); clear e; intros t0 e H1 H3.
      exists (t0 --> t); split.
      rewrite stype_w_free_arr; split; auto; apply (H1 pos0).
      exists e; split.
      apply ita_abs; auto.
      intros p; apply (H1 (pos_nxt _ p)).
     
      destruct 1.

      destruct (proj1 IHa) as (ta & Ha2 & Hta).
      generalize (proj2 IHf Hf1 ((ta::nil) --> µ x)); intros H.
      inst H.
      rewrite stype_w_free_arr; split.
      rewrite itype_w_free_sg; auto.
      apply stype_w_free_var.
      exists (µ x); split.
      apply stype_w_free_var.
      apply w_free_typable_app with ta; auto.

      intros _ t Ht.
      destruct (proj1 IHa) as (h & Hh1 & Hh2).
      generalize (proj2 IHf Hf1 ((h::nil) --> t)); intros H.
      inst H.
      rewrite stype_w_free_arr; split; auto.
      rewrite itype_w_free_sg; auto.
      apply w_free_typable_app with h; auto.
      
      exists (µ x); split.
      apply stype_w_free_var.
      apply w_free_typable_var, stype_w_free_var.
      
      intros _.
      apply w_free_typable_var.
    Qed.
    
    (* Any normal term can be typed in DOmega without using Omega *)

    Theorem ut_normal_typable_wo_omega n : @ut_normal n inc1 @typable_wo_omega n.
    Proof.
      intros f H.
      destruct (ut_normal_typable_w_free_rec H) as [ (t & ? & e & ? & ?) _ ].
      exists e, (t::nil); split; auto; split; auto.
      apply itype_w_free_sg; auto.
    Qed.

    Corollary ut_beta_eta_normalizable_typable_polarity_free n :
          @ut_beta_eta_normalizable n inc1 @typable_polarity_free n.
    Proof.
      intros f (g & H1 & H2).
      apply ut_beta_eta_normal_beta_normal, 
            ut_normal_typable_wo_omega,
            typable_wo_omega_typable_polarity_free
        in H2.
      revert H1 H2.
      apply typable_pf_beta_eta_rt_rev.
    Qed.

  End typing_of_normal_terms.
  
End system_D_omega.
