(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith List Relations.

Require Import utils cr pos.
Require Import uterm uterm_beta.

Set Implicit Arguments.

Local Reserved Notation " x '-r->' y " (at level 71).

Inductive ut_rho : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_rho_var   : forall n (p : pos n),                                      £ p -r-> £ p
  | in_ut_rho_redex : forall n f g (a b : uterm n), f -r-> g -> a -r-> b -> \\ f @ a -r-> g[b] 
  | in_ut_rho_app   : forall n f g (a b : uterm n), f -r-> g -> a -r-> b ->    f @ a -r-> g @ b 
  | in_ut_rho_abs   : forall n (f g : uterm (S n)), f -r-> g ->                 \\ f -r-> \\ g
where  " x '-r->' y " := (@ut_rho _ x y).

Definition ut_rho_rt n := clos_refl_trans _ (@ut_rho n).

Local Notation " x '-r>>' y " := (@ut_rho_rt _ x y) (at level 71).

Fact in_ut_rho_refl n (f : uterm n) : f -r-> f.
Proof.
  induction f.
  apply in_ut_rho_abs; auto.
  apply in_ut_rho_app; auto.
  apply in_ut_rho_var.
Qed.

Fact ut_beta_rho n (f g : uterm n) : f -b-> g -> f -r-> g.
Proof.
  induction 1.
  apply in_ut_rho_redex; apply in_ut_rho_refl.
  apply in_ut_rho_app; auto; apply in_ut_rho_refl.
  apply in_ut_rho_app; auto; apply in_ut_rho_refl.
  apply in_ut_rho_abs; auto.
Qed.

Fact ut_rho_beta_rt n (f g : uterm n) : f -r-> g -> f -b>> g.
Proof.
  induction 1 as [ n f  
                 | n f g a b H1 IH1 H2 IH2 
                 | n f g a b H1 IH1 H2 IH2
                 | n f g H IH
                 ].
  constructor 2.
  constructor 3 with (\\ g @  b).
  apply ut_beta_rt_app; auto; apply ut_beta_rt_abs; auto.
  constructor 1; constructor 1.
  apply ut_beta_rt_app; auto.
  apply ut_beta_rt_abs; auto.
Qed.

Fact ut_beta_rho_rt_eq n (f g : uterm n) : f -b>> g <-> f -r>> g.
Proof.
  split.
  induction 1.
  constructor 1.
  apply ut_beta_rho; auto.
  constructor 2.
  constructor 3 with y; auto.
  induction 1.
  apply ut_rho_beta_rt; auto.
  constructor 2.
  constructor 3 with y; auto.
Qed.

Section ut_rho_inv.

  Let ut_rho_inv n (u v : uterm n) : u -r-> v -> (exists p, u = £ p /\ v = £ p) 
                                              \/ (exists f g a b, u = \\ f @ a /\ v = g[b] /\ f -r-> g /\ a -r-> b)
                                              \/ (exists f g a b, u = f @ a /\ v = g @ b /\ f -r-> g /\ a -r-> b)
                                              \/ (exists f g, u = \\ f /\ v = \\ g /\ f -r-> g).
  Proof.
    destruct 1; [ left | do 1 right; left | do 2 right; left | do 3 right ].
    exists p; auto.
    exists f, g, a, b; auto.
    exists f, g, a, b; auto.
    exists f, g; auto.
  Qed.

  Fact ut_rho_var_inv n (v : uterm n) p : £ p -r-> v -> v = £ p.
  Proof.
    intros H; apply ut_rho_inv in H.
    destruct H as [ (q & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?)
                  | (f & g & ? & ? & ?)
                  ] ] ]; try discriminate H. 
    subst; auto.
  Qed.

  Fact ut_rho_abs_inv n (v : uterm n) u : \\ u -r-> v -> exists g, v = \\ g /\ u -r-> g.
  Proof. 
    intros H; apply ut_rho_inv in H.
    destruct H as [ (q & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?)
                  | (f & g & ? & ? & ?)
                  ] ] ]; try discriminate H.
    apply ut_abs_inj in H.
    subst; exists g; auto.
  Qed.
 
  Fact ut_rho_app_inv n (v h u : uterm n) : h @ u -r-> v 
                                         -> (exists f g b, h = \\ f /\ v = g[b] /\ f -r-> g /\ u -r-> b) 
                                         \/ (exists g b, v = g @ b /\ h -r-> g /\ u -r-> b).
  Proof.
    intros H; apply ut_rho_inv in H.
    destruct H as [ (q & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?) 
                | [ (f & g & a & b & ? & ? & ? & ?)
                  | (f & g & ? & ? & ?)
                  ] ] ]; try discriminate H; [ left | right ];
      apply ut_app_inj in H; destruct H; subst.
    exists f, g, b; auto.
    exists g, b; auto.
  Qed.

  Fact ut_rho_redex_inv n f (a u : uterm n) : 
          \\ f @ u -r-> a 
       -> exists g v, f -r-> g /\ u -r-> v /\ (a = g[v] \/ a = \\ g @ v).
  Proof.
    intros H.
    apply ut_rho_app_inv in H.
    destruct H as [ ( g & h & v & H0 & H1 & H2 & H3 ) 
                  | ( g & v & H0 & H1 & H2 ) ]; subst.
    apply ut_abs_inj in H0; subst.
    exists h, v; auto.
    apply ut_rho_abs_inv in H1.
    destruct H1 as ( h & ? & H1 ); subst.
    exists h, v; auto.
  Qed.
  
  Fact ut_rho_app_neutral_inv n f (a u : uterm n) : 
          f @ u -r-> a 
       -> ut_neutral f 
       -> exists g v, a = g @ v /\ f -r-> g /\ u -r-> v.
  Proof.
    intros H1 H.
    apply ut_rho_app_inv in H1.
    destruct H1 as [ ( g & h & v & H0 & H1 & H2 & H3 ) 
                   | ( g & v & H0 & H1 & H2 ) ]; subst.
    contradict H.
    exists g, v; auto.
  Qed.

End ut_rho_inv.

Section ut_rho_subst.

  Let ut_rho_ren n (f g : uterm n) m (r : pos_map n m) : f -r-> g -> ut_ren f r -r-> ut_ren g r.
  Proof.
    intros H; revert H m r.
    induction 1 as [ n f  
                   | n f g a b H1 IH1 H2 IH2 
                   | n f g a b H1 IH1 H2 IH2
                   | n f g H IH
                   ]; intros m r; simpl. 
    apply in_ut_rho_var.
    rewrite ut_ren_reduce.
    apply in_ut_rho_redex; auto.
    apply in_ut_rho_app; auto.
    apply in_ut_rho_abs; auto.
  Qed.

  Let ut_rho_lts_lift n m (r s : uts_t n m) :
      (forall p, r p -r-> s p) -> forall p, uts_lift r p -r-> uts_lift s p.
  Proof.
    intros H p.
    unfold uts_lift.
    invert pos p; simpl.
    apply in_ut_rho_var.
    unfold ut_lift.
    apply ut_rho_ren, H.
  Qed.

  Fact ut_rho_subst n (f g : uterm n) m (r s : uts_t n m) :
        (forall p, r p -r-> s p) -> f -r-> g -> ut_subst f r -r-> ut_subst g s.
  Proof.
    intros H1 H2; revert H2 m r s H1.
    induction 1 as [ n f  
                   | n f g a b H1 IH1 H2 IH2 
                   | n f g a b H1 IH1 H2 IH2
                   | n f g H IH
                   ]; intros m r s Hrs; simpl.
    apply Hrs.
    rewrite ut_subst_reduce.
    apply in_ut_rho_redex.
    apply IH1.
    apply ut_rho_lts_lift; auto.
    apply IH2; auto.
    apply in_ut_rho_app; auto.
    apply in_ut_rho_abs.
    apply IH.
    apply ut_rho_lts_lift; auto.
  Qed.

End ut_rho_subst.

Fact ut_rho_reduce n f g (a b : uterm n) : f -r-> g -> a -r-> b -> f[a] -r-> g[b].
Proof.
  intros; apply ut_rho_subst; auto.
  intros p; unfold uts_first; invert pos p; auto.
  apply in_ut_rho_var.
Qed.

(* Notice the us of a custom induction principle for uterm that
   distinguishes between redex and neutral app *)

Theorem ut_rho_diamond n : diamond (@ut_rho n).
Proof.
  intros f a b; revert a b.
  induction f as [ n f IHf 
                 | n f a Hf IHf IHa 
                 | n f a IHf IHa 
                 | n p ] using ut_custom_rect; 
    intros u v Hu Hv.
  
  (* one case for ut_abs _ *)

  destruct (ut_rho_abs_inv Hu) as (u' & ? & Hu').
  destruct (ut_rho_abs_inv Hv) as (v' & ? & Hv').
  subst.
  destruct (IHf _ _ Hu' Hv') as (g & ? & ?).
  exists (\\ g); split; apply in_ut_rho_abs; auto.
  
  (* one case for ut_app f _ with a neutral f *)
  
  destruct ut_rho_app_neutral_inv with (1 := Hu) (2 := Hf) as (f1 & a1 & ? & G0 & G1).
  destruct ut_rho_app_neutral_inv with (1 := Hv) (2 := Hf) as (f2 & a2 & ? & G2 & G3).
  subst.
  destruct (IHf _ _ G0 G2) as (g & ? & ?).
  destruct (IHa _ _ G1 G3) as (b & ? & ?).
  exists (g @ b); split; apply in_ut_rho_app; auto.
  
  (* four cases for ut_app (ut_abs _) _ *)
  
  destruct ut_rho_redex_inv with (1 := Hu) as (f1 & a1 & H1 & H2 & H3).
  destruct ut_rho_redex_inv with (1 := Hv) as (f2 & a2 & H4 & H5 & H6).
  destruct (IHf _ _ H1 H4) as (g & ? & ?).
  destruct (IHa _ _ H2 H5) as (b & ? & ?).
  exists (g[b]); destruct H3; destruct H6; subst; split;
   ( apply ut_rho_reduce || apply in_ut_rho_redex); auto.
  
  (* one case for ut_var _ *)
  
  apply ut_rho_var_inv in Hu.
  apply ut_rho_var_inv in Hv.
  subst; exists (£ p); split; apply in_ut_rho_var.
Qed.

Theorem ut_beta_confluent n : confluent (@ut_beta n).
Proof.
  generalize (square_crt (@ut_rho_diamond n)).
  apply diamond_eq.
  split; apply ut_beta_rho_rt_eq.
Qed.

Theorem ut_beta_rt_cr n : church_rosser (@ut_beta_rt n).
Proof.
  apply ut_beta_confluent.
Qed.

