(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import utils pos.
Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_head uterm_normal uterm_solvable uterm_head_sem.

Set Implicit Arguments.

Definition ut_true n : uterm n.
Proof.
  do 2 apply ut_abs.
  apply ut_var, pos_nxt, pos_fst.
Defined.

Definition ut_false n : uterm n.
Proof.
  do 2 apply ut_abs.
  apply ut_var, pos_fst.
Defined.

Definition ut_ite n : uterm n.
Proof.
  do 3 apply ut_abs.
  apply ut_app.
  apply ut_app.
  apply ut_var, pos_nxt, pos_nxt, pos_fst.
  apply ut_var, pos_nxt, pos_fst.
  apply ut_var, pos_fst.
Defined.

Fact ut_true_subst n k (s : uts_t n k) : ut_subst (ut_true n) s = ut_true _.
Proof. auto. Qed.

Fact ut_false_subst n k (s : uts_t n k) : ut_subst (ut_false n) s = ut_false _.
Proof. auto. Qed.

Fact ut_ite_subst n k (s : uts_t n k) : ut_subst (ut_ite n) s = ut_ite _.
Proof. auto. Qed.

Fact ut_true_normal n : ut_normal (ut_true n).       Proof. ut_normal_tac. Qed.
Fact ut_false_normal n : ut_normal (ut_false n).     Proof. ut_normal_tac. Qed.
Fact ut_ite_normal n : ut_normal (ut_ite n).         Proof. ut_normal_tac. Qed.

Fact ut_true_solvable n : ut_solvable (ut_true n).
Proof. apply ut_normal_solvable, ut_true_normal. Qed.

Fact ut_false_solvable n : ut_solvable (ut_false n).
Proof. apply ut_normal_solvable, ut_false_normal. Qed.

Fact ut_true_false_beta_neq n : ~ ut_true _ ~b ut_false n.
Proof.
  intros H.
  apply ut_beta_eq_normal in H.
  discriminate H.
  apply ut_true_normal.
  apply ut_false_normal.
Qed.

Proposition ut_true_prop n x y : @ut_true n @ x @ y -w>> x.
Proof.
  unfold ut_true.  
  repeat (ut_whead_rt_step; ut_simpl).
  constructor 2.
Qed.

Proposition ut_false_prop n x y : @ut_false n @ x @ y -w>> y.
Proof.
  unfold ut_false.
  repeat (ut_whead_rt_step; ut_simpl).
  constructor 2.
Qed.

(* Lemma 2.12 p 35, require the characterization of solvability by head_sn *)

Fact ut_beta_eq_true_whead n (b t0 t1 : uterm n) : 
  b ~b ut_true _ -> b @ t0 @ t1 -w>> t0.
Proof.
  intros H.
  destruct (@ut_beta_rt_whead_rt n 2 b (pos_nxt _ (pos_fst _)) nil H) as (m & Hm1 & Hm2).
  destruct m; try inversion Hm1; clear Hm1.
  simpl in Hm2.
  set (s := fun p : pos 2 => match p with
                               | pos_fst  _ => ut_lift (ut_lift t1)
                               | _          => ut_lift (ut_lift t0)
                             end).
  specialize (Hm2 s); unfold s in Hm2; simpl in Hm2; clear s.
  do 2 apply ut_whead_rt_inv_lift.
  apply Hm2.
Qed.

Fact ut_beta_eq_false_whead n (b t0 t1 : uterm n) : 
  b ~b ut_false _ -> b @ t0 @ t1 -w>> t1.
Proof.
  intros H.
  destruct (@ut_beta_rt_whead_rt n 2 b (pos_fst _) nil H) as (m & Hm1 & Hm2).
  destruct m; try inversion Hm1; clear Hm1.
  simpl in Hm2.
  set (s := fun p : pos 2 => match p with
                               | pos_fst  _ => ut_lift (ut_lift t1)
                               | _          => ut_lift (ut_lift t0)
                             end).
  specialize (Hm2 s); unfold s in Hm2; simpl in Hm2; clear s.
  do 2 apply ut_whead_rt_inv_lift.
  apply Hm2.
Qed.

Proposition ut_ite_prop n b x y : @ut_ite n @ b @ x @ y -w>> b @ x @ y.
Proof.
  unfold ut_ite.
  repeat (ut_whead_rt_step; ut_simpl).
  constructor 2.
Qed.

Proposition ut_ite_true_prop n x y : @ut_ite n @ ut_true _ @ x @ y -w>> x.
Proof.
  apply ut_whead_rt_trans with (1 := ut_ite_prop _ _ _).
  apply ut_true_prop.
Qed.

Proposition ut_ite_true_beta_prop n b x y : 
   b -b>> ut_true n -> ut_ite _ @ b @ x @ y -b>> x.
Proof.
  intros H.
  apply ut_beta_rt_trans with (ut_ite _ @ ut_true _ @ x @ y).
  apply ut_beta_rt_lft, ut_beta_rt_lft, ut_beta_rt_rt, H.
  apply ut_whead_rt_beta_rt, ut_ite_true_prop.
Qed.

Proposition ut_ite_false_prop n x y : @ut_ite n @ ut_false _ @ x @ y -w>> y.
Proof.
  apply ut_whead_rt_trans with (1 := ut_ite_prop _ _ _).
  apply ut_false_prop.
Qed.

Proposition ut_ite_false_beta_prop n b x y : 
    b -b>> ut_false n -> ut_ite _ @ b @ x @ y -b>> y.
Proof.
  intros H.
  apply ut_beta_rt_trans with (ut_ite _ @ ut_false _ @ x @ y).
  apply ut_beta_rt_lft, ut_beta_rt_lft, ut_beta_rt_rt, H.
  apply ut_whead_rt_beta_rt, ut_ite_false_prop.
Qed.

Definition ut_pair n x y := \\ (£ (pos_fst n) @ ut_lift x @ ut_lift y).
Definition ut_fst n : uterm n := \\ (£ (pos_fst _) @ ut_true _).
Definition ut_snd n : uterm n := \\ (£ (pos_fst _) @ ut_false _).

Fact ut_pair_subst n m u v (s : uts_t n m) : ut_subst (@ut_pair n u v) s = ut_pair (ut_subst u s) (ut_subst v s).
Proof.
  unfold ut_pair; ut_simpl; do 2 f_equal; [ f_equal | idtac ];
  apply ut_subst_ext; intro; rewrite ut_ren_subst_eq; auto.
Qed.

Fact ut_pair_lift n u v : ut_lift (@ut_pair n u v) = ut_pair (ut_lift u) (ut_lift v).
Proof.
  rewrite ut_lift_subst_eq, ut_pair_subst; f_equal; symmetry; apply ut_lift_subst_eq.
Qed.

Fact ut_fst_subst n m (s : uts_t n m) : ut_subst (ut_fst _) s = ut_fst _.
Proof. auto. Qed.

Fact ut_snd_subst n m (s : uts_t n m) : ut_subst (ut_snd _) s = ut_snd _.
Proof. auto. Qed.

Fact ut_fst_reduce n u : ut_fst n @ u -w>> u @ ut_true _.
Proof.
  unfold ut_fst.
  ut_whead_rt_step; ut_simpl.
  constructor 2.
Qed.

Fact ut_snd_reduce n u : ut_snd n @ u -w>> u @ ut_false _.
Proof.
  unfold ut_snd.
  ut_whead_rt_step; ut_simpl.
  constructor 2.
Qed.

Fact ut_fst_solvable n u : ut_solvable (ut_fst n @ u) -> ut_solvable u.
Proof.
  intros H.
  generalize (@ut_fst_reduce n u); intros H1.
  apply ut_whead_rt_beta_rt in H1.
  apply ut_beta_rt_solvable in H1.
  apply H1 in H; clear H1.
  revert H; apply ut_app_solvable.
Qed.

Fact ut_snd_solvable n u : ut_solvable (ut_snd n @ u) -> ut_solvable u.
Proof.
  intros H.
  generalize (@ut_snd_reduce n u); intros H1.
  apply ut_whead_rt_beta_rt in H1.
  apply ut_beta_rt_solvable in H1.
  apply H1 in H; clear H1.
  revert H; apply ut_app_solvable.
Qed.

Fact ut_fst_pair_prop n x y : @ut_fst n @ ut_pair x y -w>> x.
Proof.
  unfold ut_fst, ut_pair.
  repeat (ut_whead_rt_step; ut_simpl).
  constructor 2.
Qed.

Fact ut_snd_pair_prop n x y : @ut_snd n @ ut_pair x y -w>> y.
Proof.
  unfold ut_snd, ut_pair.
  repeat (ut_whead_rt_step; ut_simpl).
  constructor 2.
Qed.

Fact ut_pair_normal n x y : ut_normal x -> ut_normal y -> ut_normal (@ut_pair n x y).
Proof.
  intros Hx Hy.
  unfold ut_pair.
  apply ut_abs_normal.
  apply ut_app_normal; simpl; auto.
  apply ut_app_normal; simpl; auto.
  apply ut_var_normal.
  apply ut_lift_normal; auto.
  apply ut_lift_normal; auto.
Qed.
