(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Omega List Relations.

Require Import utils pos cr finitary_rel.
Require Import uterm.

Set Implicit Arguments.

Reserved Notation " x '-e->' y " (at level 71).

(* Beware that ut_eta x y is denote y -e-> x (reversed) *)

Inductive ut_eta : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_eta_redex  : forall n (f : uterm n) ,                    \\ (ut_lift f @ £ pos0) -e-> f  
  | in_ut_eta_lft    : forall n (x y k : uterm n),     x -e-> y ->                   x @ k -e-> y @ k
  | in_ut_eta_rt     : forall n (x y k : uterm n),     x -e-> y ->                   k @ x -e-> k @ y
  | in_ut_eta_abs    : forall n (x y : uterm (S n)),   x -e-> y ->                   \\ x  -e-> \\ y
where  " x '-e->' y " := (@ut_eta _ x y).

Definition ut_eta_rt n := clos_refl_trans _ (@ut_eta n).

Infix "-e>>" := (@ut_eta_rt _) (at level 71).

Fact ut_eta_rt_refl n (f : uterm n) : f -e>> f.
Proof. constructor 2. Qed.

Fact ut_eq_eta_rt n (f g : uterm n) : f = g -> f -e>> g.
Proof. intros []; constructor 2. Qed.

Fact ut_eta_rt_trans n (f g h : uterm n) : f -e>> g -> g -e>> h -> f -e>> h.
Proof. intros; constructor 3 with g; auto. Qed.

Fact ut_eta_eta_rt n (f g : uterm n) : f -e-> g -> f -e>> g.
Proof. intros; constructor 1; auto. Qed.

Fact ut_eta_rt_inv n (f g : uterm n) : f -e>> g -> f = g \/ exists h, f -e-> h /\ h -e>> g.
Proof. apply clos_refl_trans_step. Qed.

Fact ut_eta_rt_redex n f (h : uterm n) : f -e>> h -> \\ (ut_lift f @ £ pos0) -e>> h.
Proof.
  intros H.
  apply ut_eta_rt_trans with (2 := H).
  constructor 1; apply in_ut_eta_redex.
Qed.

Fact ut_eta_rt_lft n (f g a : uterm n) : f -e>> g -> f @ a -e>> g @ a.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @ a).
  intros ? ?; apply in_ut_eta_lft.
Qed.

Fact ut_eta_rt_rt n (f a b : uterm n) : a -e>> b -> f @ a -e>> f @ b.
Proof.
  apply clos_refl_trans_map with (f := fun x => f @ x).
  intros ? ?; apply in_ut_eta_rt.
Qed.

Fact ut_eta_rt_app n (f g a b : uterm n) : f -e>> g -> a -e>> b -> f @ a -e>> g @ b.
Proof.
  intros H1 H2.
  constructor 3 with (f @ b).
  apply ut_eta_rt_rt; auto.
  apply ut_eta_rt_lft; auto.
Qed.

Fact ut_eta_rt_abs n (f g : uterm (S n)) : f -e>> g -> \\ f -e>> \\ g.
Proof.
  apply clos_refl_trans_map with (f := @ut_abs _).
  intros ? ?; apply in_ut_eta_abs.
Qed.

Fact ut_eta_subst n (f g : uterm n) m (s : uts_t n m) : f -e-> g -> ut_subst f s -e-> ut_subst g s.
Proof.
  intros H; revert H m s.
  induction 1 as [ n f' a' | n f' h' a' H1 IH | n f' a' b' H1 IH | n f' h' H IH ]; intros m s; simpl.
  generalize (in_ut_eta_redex (ut_subst f' s)); eqgoal.
  f_equal; f_equal.
  rewrite <- ut_lift_subst_2; reflexivity.
  apply in_ut_eta_lft; auto.
  apply in_ut_eta_rt; auto.
  apply in_ut_eta_abs; auto.
Qed.

Fact ut_eta_rt_subst n (f g : uterm n) m (s : uts_t n m) : f -e>> g -> ut_subst f s -e>> ut_subst g s.
Proof.
  apply clos_refl_trans_map with (f := fun x => ut_subst x s).
  intros ? ?; apply ut_eta_subst.
Qed.

Fact ut_eta_ren n (f g : uterm n) m (r : pos_map n m) : f -e-> g -> ut_ren f r -e-> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_eta_subst. Qed.

Fact ut_eta_rt_ren n (f g : uterm n) m (r : pos_map n m) : f -e>> g -> ut_ren f r -e>> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_eta_rt_subst. Qed.

Fact ut_eta_lift n (f g : uterm n) : f -e-> g -> ut_lift f -e-> ut_lift g.
Proof. apply ut_eta_ren. Qed.

Fact ut_eta_rt_lift n (f g : uterm n) : f -e>> g -> ut_lift f -e>> ut_lift g.
Proof. apply ut_eta_rt_ren. Qed.

Fact ut_eta_rt_subst_any n f m (s t : uts_t n m) : (forall p, s p -e>> t p) -> ut_subst f s -e>> ut_subst f t.
Proof.
  revert m s t.
  induction f as [ n f IHf | n a IHa b IHb | n v ]; intros m s t H; simpl.
  apply ut_eta_rt_abs, IHf.
  intros p; pos_inv p; simpl.
  apply ut_eta_rt_refl.
  apply ut_eta_rt_ren; auto.
  constructor 3 with (ut_subst a s @ ut_subst b t).
  apply ut_eta_rt_rt, IHb; auto.
  apply ut_eta_rt_lft, IHa; auto.
  apply H.
Qed.

Fact ut_eta_rt_reduce n f (a b : uterm n) : a -e>> b -> f[a] -e>> f[b].
Proof.
  intros H; apply ut_eta_rt_subst_any.
  intros p; pos_inv p.
  apply H.
  apply ut_eta_rt_refl.
Qed.

Fact ut_eta_occ n (f g : uterm n) p : f -e-> g -> ut_occ f p = ut_occ g p.
Proof.
  induction 1; simpl; auto; try (f_equal; auto; fail).
  rewrite ut_occ_lift_nxt; omega.
Qed.

Section ut_eta_inv.

  Let ut_eta_inv n (u v : uterm n) : u -e-> v -> (exists f,     u = \\ (ut_lift f @ £ pos0) /\ v = f)
                                              \/ (exists f g a, u = f @ a    /\ v = g @ a /\ f -e-> g)
                                              \/ (exists f a b, u = f @ a    /\ v = f @ b /\ a -e-> b)
                                              \/ (exists f g,   u = \\ f     /\ v = \\ g  /\ f -e-> g).
  Proof.
    destruct 1 as [ n f | n f g a H | n a b f H | n f g H ].
    do 0 right; left; exists f; auto.
    do 1 right; left; exists f, g, a; auto.
    do 2 right; left; exists f, a, b; auto.
    do 3 right; exists f, g; auto.
  Qed.

  Fact ut_eta_var_inv n (p : pos n) h : ~ £ p -e-> h.
  Proof. inversion 1. Qed.
  
  Fact ut_eta_abs_inv n u (v : uterm n) : \\ u -e-> v ->  u = ut_lift v @ £ pos0 
                                                      \/ (exists w, v = \\ w /\ u -e-> w).
  Proof.
    intros H; apply ut_eta_inv in H.
    destruct H as [ (f & H1 & H2) 
                | [ (f & g & a & H1 & H2 & H3) 
                | [ (f & a & b & H1 & H2 & H3) 
                |   (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1.
    left; subst; apply ut_abs_inj in H1; auto.
    right; apply ut_abs_inj in H1; subst u v; exists g; auto.
  Qed.

  Fact ut_eta_app_inv n (h u v : uterm n) : 
           h @ u -e-> v 
       -> (exists g, v = g @ u /\ h -e-> g)
       \/ (exists w, v = h @ w /\ u -e-> w).
  Proof.
    intros H; apply ut_eta_inv in H.
    destruct H as [ (f & H1 & H2) 
                | [ (f & g & a & H1 & H2 & H3) 
                | [ (f & a & b & H1 & H2 & H3) 
                |   (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1;
    apply ut_app_inj in H1; destruct H1; subst h u v.
    left; exists g; auto.
    right; exists b; auto.
  Qed.

  Fact ut_rev_eta_abs_inv n (u : uterm n) v : 
    u -e-> \\ v -> (exists f', u = \\ f' /\ f' -e-> v)
                \/ u = \\ (ut_lift (\\v) @ £ pos0).
  Proof.
    intros H.
    apply ut_eta_inv in H.
    destruct H as [ (f & H1 & H2) 
                | [ (f & g & a & H1 & H2 & H3) 
                | [ (f & a & b & H1 & H2 & H3) 
                |   (f & g & H1 & H2 & H3) ] ] ]; subst.
    right; auto.
    discriminate H2.
    discriminate H2.
    apply ut_abs_inj in H2; subst.
    left; exists f; auto.
  Qed.

  Fact ut_rev_eta_app_inv n (u v w : uterm n) :
    u -e-> v @ w -> (exists v', u = v' @ w /\ v' -e-> v)
                 \/ (exists w', u = v @ w' /\ w' -e-> w)
                 \/ u = \\ ((ut_lift v @ ut_lift w) @ £ pos0). 
  Proof.
    intros H.
    apply ut_eta_inv in H.
    destruct H as [ (f & H1 & H2) 
                | [ (f & g & a & H1 & H2 & H3) 
                | [ (f & a & b & H1 & H2 & H3) 
                |   (f & g & H1 & H2 & H3) ] ] ]; subst.
    right; right; auto.
    apply ut_app_inj in H2; destruct H2; subst g a.
    left; exists f; auto.
    apply ut_app_inj in H2; destruct H2; subst f b.
    right; left; exists a; auto.
    discriminate H2.
  Qed.

  Fact ut_rev_eta_var_inv n u (p : pos n) :
    u -e-> £ p -> u = \\ (£ (pos_nxt _ p) @ £ pos0).
  Proof.
    intros H.
    apply ut_eta_inv in H.
    destruct H as [ (f & H1 & H2) 
                | [ (f & g & a & H1 & H2 & H3) 
                | [ (f & a & b & H1 & H2 & H3) 
                |   (f & g & H1 & H2 & H3) ] ] ]; subst.
    auto.
    discriminate H2.
    discriminate H2.
    discriminate H2.
  Qed.

End ut_eta_inv.

Section ut_eta_ren_rel_square.

  Fact ut_eta_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_eta _) (ut_ren_rel R) (ut_ren_rel R) (@ut_eta _).
  Proof.
    intros f g u H1.
    revert H1 m R u.
    induction 1 as [ n f | n f g a H1 IH1 | n a b f H1 IH1 | n f g H1 IH1 ];
      intros m R u H.

    apply ut_ren_rel_abs_inv in H.
    destruct H as (h & ? & H); subst.
    apply ut_ren_rel_app_inv in H.
    destruct H as (g & b & ? & H1 & H2); subst.
    apply ut_ren_rel_var_inv in H2.
    destruct H2 as (p & ? & H2); subst.
    pos_inv p.
    2: destruct H2.
    clear H2.

    (* here we use the compositionality of ut_ren_rel *)

    assert (f ~[(fun x y => y = pos_nxt _ x) o ren_rel_lift R]~ g) as H2.
      apply ut_ren_rel_comp.
      exists (ut_lift f); split; auto.
      apply ut_ren_rel_eq_ren.
    assert (f ~[R o (fun z y => y = pos_nxt _ z )]~ g) as H3.
      revert H2; apply ut_ren_rel_mono.
      intros p q (r & ? & H3); subst; auto.
      pos_inv q.
      destruct H3.
      cbv in H3.
      exists q; auto.
    apply ut_ren_rel_comp in H3.
    destruct H3 as (h & H3 & H4).
    exists h; split; auto.
    apply ut_ren_rel_fun_eq in H4; subst.
    constructor 1.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H2 & H3); subst.
    destruct (IH1 _ _ _ H2) as (v & ? & ?).
    exists (v @ b); split.
    constructor; auto.
    constructor 2; auto.

    apply ut_ren_rel_app_inv in H.
    destruct H as (g & c & ? & H2 & H3); subst.
    destruct (IH1 _ _ _ H3) as (v & ? & ?).
    exists (g @ v); split.
    constructor; auto.
    constructor 3; auto.

    apply ut_ren_rel_abs_inv in H.
    destruct H as (h & ? & H); subst.
    destruct (IH1 _ _ _ H) as (v & ? & ?).
    exists (\\ v); split.
    constructor; auto.
    constructor; auto.
  Qed.

End ut_eta_ren_rel_square.

Corollary ut_eta_ren_inv n m f g (r : pos_map n m) : ut_ren f r -e-> g -> exists g', g = ut_ren g' r /\ f -e-> g'.
Proof.
  intros H1.
  generalize (ut_ren_rel_eq_ren f r); intros H2.
  apply ut_ren_rel_sym in H2.
  destruct (ut_eta_ren_rel_comm H1 H2) as (g' & H3 & H4).
  exists g'; split; auto.
  apply ut_ren_rel_fun_eq, ut_ren_rel_sym; auto.
Qed.
  
Corollary ut_eta_lift_inv n (f : uterm n) g : ut_lift f -e-> g -> exists g', g = ut_lift g' /\ f -e-> g'.
Proof.
  apply ut_eta_ren_inv.
Qed.  

(* The proof of SN of -e-> is trivial if you do it on the size of the lambda terms
   It does not work if you do it on the number of eta redexes ...
   
       \\y.(u @ (\\x.(y @ x))) -e-> \\y.(u @ y) -e-> u

   but the first two terms both have 1 eta redex. So the number of eta redexes
   does not strictly decrease under eta reduction even though you must
   finally reach an eta-redex free lambda term
*) 

Fact ut_eta_size n (g f : uterm n) : f -e-> g -> ut_size g < ut_size f. 
Proof.
  induction 1; simpl; try omega.
  unfold ut_lift; rewrite ut_ren_size; omega.
Qed.

Notation ut_eta_sn := (Acc (inv (@ut_eta _))).

Require Import Wellfounded.

Theorem ut_eta_strongly_normalizing n (f : uterm n) : ut_eta_sn f.
Proof.
  revert f.
  apply wf_incl with (1 := @ut_eta_size n).
  apply wf_inverse_image, lt_wf.
Qed.

Fact ut_eta_finite n (f : uterm n) : { l | forall g, f -e-> g <-> In g l }.
Proof.
  induction f as [ n f IHf | n a IHa b IHb | n v ].
  destruct IHf as (l & Hl).
  destruct f as [ f' | h a | v ].
  
  exists (map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  discriminate H.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.  
  rewrite in_map_iff.
  intros (h & ? & Hh); subst.
  constructor 4; apply Hl; auto.
  
  destruct a as [ a | a b | v ].
  
  exists  (map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  discriminate H.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.  
  rewrite in_map_iff.
  intros (u & ? & Hu); subst.
  constructor 4; apply Hl; auto.
  
  exists  (map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  discriminate H.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.  
  rewrite in_map_iff.
  intros (u & ? & Hu); subst.
  constructor 4; apply Hl; auto.
  
  pos_inv v.
  
  destruct (eq_nat_dec (ut_occ h pos0) 0) as [ E | C ].
  apply ut_find_lift in E.
  destruct E as (g & ?); subst.
  exists (g::map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  apply ut_app_inj, proj1, ut_lift_inj in H; subst.
  left; auto.
  right; apply in_map_iff.
  exists w; split; auto; apply Hl; auto.
  intros [ [] | ].
  constructor 1.
  rewrite in_map_iff.
  intros (u & ? & Hu); subst.
  constructor 4; apply Hl; auto.
  
  exists (map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  apply ut_app_inj, proj1 in H; subst h.
  rewrite ut_occ_lift_fst in C; contradict C; auto.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.  
  rewrite in_map_iff.
  intros (u & ? & Hu); subst.
  constructor 4; apply Hl; auto.
  
  exists (map (fun x => \\ x) l); split.
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  discriminate H.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.  
  rewrite in_map_iff.
  intros (u & ? & Hu); subst.
  constructor 4; apply Hl; auto.
  
  exists nil.
  intros g; split.
  2: intros [].
  intros H.
  apply ut_eta_abs_inv in H.
  destruct H as [ H | (w & ? & Hw) ]; subst.
  discriminate H.
  apply ut_eta_var_inv in Hw; destruct Hw.
  
  destruct IHa as (la & Hla).
  destruct IHb as (lb & Hlb).
  exists (map (fun x => x @ b) la ++ map (fun x => a @ x) lb).
  intros g; split; intros Hg.
  apply in_or_app.
  apply ut_eta_app_inv in Hg.
  destruct Hg as [ (w & ? & Hw) | (w & ? & Hw) ]; subst.
  left; apply in_map_iff; exists w; split; auto; apply Hla; auto.
  right; apply in_map_iff; exists w; split; auto; apply Hlb; auto.
  apply in_app_or in Hg.
  destruct Hg as [ Hg | Hg ].
  apply in_map_iff in Hg; destruct Hg as (u & ? & Hu); subst.
  constructor 2; apply Hla; auto.
  apply in_map_iff in Hg; destruct Hg as (u & ? & Hu); subst.
  constructor 3; apply Hlb; auto.
  
  exists nil.
  split.
  2: intros [].
  intros H.
  apply ut_eta_var_inv in H; apply H.
Qed.

Definition ut_eta_normal_dec n (f : uterm n) : { g | f -e-> g } + { forall g, ~ f -e-> g }.
Proof.
  apply finitary_red_or_normal, ut_eta_finite.
Qed.

Theorem ut_eta_normal_form n (f : uterm n) : { g | f -e>> g /\ forall h, ~ g -e-> h }.
Proof.
  apply finitary_normal_form.
  apply ut_eta_finite.
  apply ut_eta_strongly_normalizing.
Qed.


