(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Bool List Relations.

Require Import utils pos.
Require Import uterm uterm_beta uterm_beta_eq.

Set Implicit Arguments.

Definition ut_beta_normal n (f : uterm n) := forall g, ~ (f -b-> g).
Notation ut_normal := ut_beta_normal.

Definition ut_beta_normalizable n (f : uterm n) := exists g, f -b>> g /\ ut_beta_normal g.
Notation ut_normalizable := ut_beta_normalizable.

Fact ut_normal_sn n f : @ut_normal n f -> ut_beta_sn f.
Proof.
  intros H; constructor 1; intros g Hg.
  apply H in Hg; destruct Hg.
Qed.

Fact ut_beta_rt_normal n (f g : uterm n) : ut_normal f -> f -b>> g -> f = g.
Proof.
  intros H1 H2.
  apply ut_beta_rt_inv in H2.
  destruct H2 as [ H2 | (h & H2 & H3) ]; auto.
  contradict H2; apply H1.
Qed. 
  
Corollary ut_beta_eq_normal n (f g : uterm n) : ut_normal f -> ut_normal g -> f ~b g -> f = g.
Proof.
  intros H1 H2 H3.
  apply ut_beta_eq_cr in H3.
  destruct H3 as (h & H3 & H4).
  generalize (ut_beta_rt_normal H1 H3) (ut_beta_rt_normal H2 H4); intros; subst; auto.
Qed.

Corollary ut_beta_normal_eq n (f g : uterm n) : ut_normal g -> f ~b g -> f -b>> g.
Proof.
  intros H1 H2.
  apply ut_beta_eq_cr in H2.
  destruct H2 as (h & H3 & H4).
  generalize (ut_beta_rt_normal H1 H4); intros; subst; auto.
Qed.

Fact ut_beta_eq_rt_normal n (f g h : uterm n) : f ~b g -> ut_normal g -> f -b>> h -> ut_normal h -> g = h.
Proof.
  intros H1 H2 H3 H4.
  apply ut_beta_eq_normal; auto.
  apply ut_beta_eq_sym.
  apply ut_beta_eq_trans with (2 := H1).
  apply ut_beta_eq_sym.
  apply ut_beta_rt_eq; auto.
Qed.

Fact ut_ren_normal n m f (s : pos_map n m) : ut_normal f -> ut_normal (ut_ren f s).
Proof.
  intros H1 g H2.
  apply ut_beta_ren_inv in H2.
  destruct H2 as (g' & H2 & H3).
  revert H3; apply H1.
Qed.

Fact ut_normal_ren n m f r : ut_normal (@ut_ren n f m r) -> ut_normal f.
Proof.
  intros H g C.
  apply (H (ut_ren g r)).
  apply ut_beta_ren;auto.
Qed.

Fact ut_lift_normal n f : ut_normal f -> ut_normal (@ut_lift n f).
Proof. apply ut_ren_normal. Qed.

Fact ut_normal_lift n f : ut_normal (@ut_lift n f) -> ut_normal f.
Proof. apply ut_normal_ren. Qed.

Fact ut_var_normal n (p : pos n) : ut_normal (£ p).
Proof. intro; apply ut_beta_var_inv. Qed.

Fact ut_abs_normal n (f : uterm (S n)) : ut_normal f <-> ut_normal (\\ f).
Proof.
  split; intros H1 g H.
  apply ut_beta_abs_inv in H.
  destruct H as (h & ? & H).
  apply (H1 _ H).
  apply in_ut_beta_abs in H; revert H; apply H1.
Qed.

Fact ut_app_normal n (f g : uterm n) : ut_neutral f -> ut_normal f -> ut_normal g -> ut_normal (f @ g).
Proof.
  intros H1 H2 H3 h H.
  apply ut_beta_app_inv in H.
  destruct H as [ (f' & G1 & G2) | [ (f' & G1 & G2) | (v & G1 & G2) ] ]; subst.
  apply H1.
  apply (H2 _ G2).
  apply (H3 _ G2).
Qed.

(* This a one step of call by value ... *)

Fixpoint ut_cbv_one n (f : uterm n) : option (uterm n) :=
  match f with 
    | \\ f  => match ut_cbv_one f with None => None | Some g => Some (\\ g) end
    | f @ a => match ut_cbv_one a with 
                 None => match f with 
                          |  \\ g => Some (g[a]) 
                          |  f    => match ut_cbv_one f with 
                                        | None => None 
                                        | Some g => Some (g @ a) 
                                     end 
                         end
                | Some b => Some (f @ b)
               end
    | £ v   => None
  end.

Fact ut_cbv_one_prop n (f : uterm n) : 
  match ut_cbv_one f with
    | None   => ut_normal f
    | Some g => f -b-> g
  end.
Proof.
  induction f as [ n f Hf | n f Hf a Ha | n v ]; simpl.
  
  destruct (ut_cbv_one f) as [ g | ].
  apply in_ut_beta_abs; auto.
  apply ut_abs_normal; auto.
  
  destruct (ut_cbv_one a) as [ b | ].
  apply in_ut_beta_rt; auto.
  destruct f as [ f | g b | v ].
  apply in_ut_beta_redex.
  destruct (ut_cbv_one (g @ b)) as [ u | ].
  apply in_ut_beta_lft; auto.
  apply ut_app_normal; auto; simpl; auto.
  destruct (ut_cbv_one (£ v)) as [ u | ].
  exfalso; revert Hf; apply ut_var_normal.
  apply ut_app_normal; auto; simpl; auto.
  
  apply ut_var_normal.
Qed.

Fixpoint ut_cbv p n : uterm n -> uterm n :=
  match p with 
    | 0   => fun x => x
    | S p => fun f => match ut_cbv_one f with None => f | Some g =>  @ut_cbv p _ g end
  end.
  
Fact ut_cbv_prop p n f : f -b>> @ut_cbv p n f.
Proof.
  revert f; induction p as [ | p IHp ]; intros f; simpl.
  apply ut_beta_rt_refl.
  generalize (ut_cbv_one_prop f).
  destruct (ut_cbv_one f) as [ g | ].
  intro; apply ut_beta_rt_trans with (2 := IHp _), ut_beta_beta_rt; auto.
  intro; apply ut_beta_rt_refl.
Qed.

(* Use this theorem to prove f -b>> g by computation. Beware that
   it is better if g is in normal form and moreover, even in that 
   case, the cbv reduction of f might not terminate *)
  
Theorem ut_beta_cbv n f g : (exists p, @ut_cbv p n f = g) -> f -b>> g.
Proof.
  intros (p & ?); subst; apply  ut_cbv_prop.
Qed.

Fact ut_normal_normalizable n (f: uterm n) : ut_normal f -> ut_normalizable f.
Proof.
  exists f; split; auto; constructor 2.
Qed.

Fact ut_abs_normalizable n (f: uterm (S n)) : ut_normalizable f <-> ut_normalizable (\\ f).
Proof.
  split.
  intros (g & H1 & H2); exists (\\ g); split.
  apply ut_beta_rt_abs; auto.
  apply ut_abs_normal; auto.
  intros (h & H1 & H2).
  apply ut_beta_rt_abs_inv in H1.
  destruct H1 as (g & ? & H1); subst.
  exists g; split; auto.
  apply ut_abs_normal; auto.
Qed.

Fact ut_var_normalizable n (p : pos n) : ut_normalizable (£ p).
Proof.
  apply ut_normal_normalizable, ut_var_normal.
Qed.

Fact ut_abs_rec_normal n m (f : uterm (m+n)) : ut_normal f -> ut_normal (\\* f).
Proof.
  revert f; induction m as [ | m IHm ]; simpl; auto; intros f Hf.
  apply IHm, ut_abs_normal, Hf.
Qed.

Fact ut_app_rev_normal n (f : uterm n) l : 
  ut_neutral f -> ut_normal f -> Forall (@ut_normal _) l -> ut_normal (f *@ l).
Proof.
  intros H1 H2.
  induction 1 as [ | a l H3 IH3 ]; simpl; auto.
  apply ut_app_normal; auto.
  destruct l; simpl; auto.
Qed.

Fact ut_beta_neutral_normal_app_rev n (f : uterm n) ll g : 
     ut_neutral f 
  -> ut_normal f 
  -> f *@ ll -b-> g 
  -> exists l a b r, ll = l++a::r /\ a -b-> b /\ g = f *@ l++b::r.
Proof.
  intros H1 H2 H3.
  apply ut_beta_neutral_app_rev_inv in H3; auto.
  destruct H3 as [ (f' & _ & H3) | (l & a & b & r & ? & ? & H3) ]; subst.
  apply H2 in H3; destruct H3.
  exists l, a, b, r; auto.
Qed.

Fact ut_beta_rt_neutral_normal_app_rev n (f : uterm n) ll g : 
     ut_neutral f 
  -> ut_normal f 
  -> f *@ ll -b>> g 
  -> exists mm,  Forall2 (@ut_beta_rt _) ll mm /\ g = f *@ mm.
Proof.
  intros H1 H2 H.
  apply clos_rt_rtn1 in H.
  induction H as [ | g h H3 H4 IH4 ].
  exists ll; split; auto.
  induction ll; constructor; auto; constructor 2.
  destruct IH4 as (mm & H5 & H6); subst.
  apply ut_beta_neutral_normal_app_rev in H3; auto.
  destruct H3 as (l & a & b & r & H3 & H6 & H7); subst.
  exists (l++b::r); split; auto.
  clear H4.
  assert (Forall2 (@ut_beta_rt _) (l++a::r) (l++b::r)) as H4.
    clear H5.
    apply Forall2_app.
    induction l; constructor; auto; constructor 2.
    constructor.
    constructor 1; auto.
    induction r; constructor; auto; constructor 2.
  generalize (Forall2_comp H5 H4).
  apply Forall2_incl.
  intros ? ? (z & ? & ?).
  apply ut_beta_rt_trans with z; auto.
Qed.
 
Fact ut_beta_rt_app_rev_neutral n (f : uterm n) l g : 
  ut_neutral f -> ut_normal f -> f *@ l -b>> g -> ut_neutral g.
Proof.
  intros Hf1 Hf2 H.
  apply ut_beta_rt_neutral_normal_app_rev in H; auto.
  destruct H as (mm & _ & ?); subst.
  destruct mm; simpl; auto.
Qed.

Fact ut_beta_rt_var_app_rev_neutral n (x : pos n) l g : £ x *@ l -b>> g -> ut_neutral g.
Proof.
  apply ut_beta_rt_app_rev_neutral.
  red; auto.
  apply ut_var_normal.
Qed.

Fact ut_normal_app n (f g : uterm n) : ut_normal (f@g) -> ut_neutral f /\ ut_normal f /\ ut_normal g.
Proof.
  intros H; repeat split.
  destruct f as [ f | | ]; simpl; auto.
  apply (H (f[g])), in_ut_beta_redex.
  intros h Hh; apply (H (h @ g)), in_ut_beta_lft, Hh.
  intros h Hh; apply (H (f @ h)), in_ut_beta_rt, Hh.
Qed.

Fact ut_normal_redex_free n f : @ut_normal n f <-> ut_redex_free f.
Proof.
  induction f as [ | n a Ha b Hb | ]; simpl.
  rewrite <- IHf, ut_abs_normal; tauto.
  rewrite <- Ha, <- Hb; split.
  apply ut_normal_app.
  intros; apply ut_app_normal; tauto.
  split; intros; auto; apply ut_var_normal.
Qed.

Fact ut_normal_b_prop n f : @ut_normal n f <-> ut_normal_b f = true.
Proof.
  rewrite ut_normal_redex_free, ut_normal_b_prop; tauto.
Qed.

Fact ut_normal_neutral n f : @ut_normal n f -> ut_neutral f -> exists l p, f = £ p *@ l.
Proof.
  induction f as [ n f IHf | n f Hf g _ | n p ]; intros H1 H2.
  contradiction H2.
  clear H2.
  apply ut_normal_app in H1.
  destruct H1 as (H1 & H2 & H3).
  destruct (Hf H2 H1) as (l & p & H); subst.
  exists (g::l), p; auto.
  exists nil, p; auto.
Qed.

Fact ut_beta_sn_app_rev n (f : uterm n) l :
     ut_neutral f
  -> ut_normal f
  -> Forall ut_beta_sn l -> ut_beta_sn (f *@ l).
Proof.
  intros Hf1 Hf2.
  induction 1 as [ | a l H1 H2 IH2 ]; simpl.
  apply ut_normal_sn; auto.
  apply ut_beta_sn_app_neutral; auto.
  intro; apply ut_beta_rt_app_rev_neutral; auto.
Qed.

Fact ut_beta_sn_app_rec n (f : uterm n) l :
     ut_neutral f
  -> ut_normal f
  -> Forall ut_beta_sn l -> ut_beta_sn (f @* l).
Proof.
  rewrite Forall_rev, ut_app_rec_rev_eq.
  apply ut_beta_sn_app_rev.
Qed.

Fact ut_beta_sn_var_app_rev n (p : pos n) l :
     Forall ut_beta_sn l -> ut_beta_sn (£ p *@ l).
Proof.
  apply ut_beta_sn_app_rev.
  red; auto.
  apply ut_var_normal.
Qed.

Fact ut_beta_sn_var_app_rec n (p : pos n) l :
     Forall ut_beta_sn l -> ut_beta_sn (£ p @* l).
Proof.
  rewrite Forall_rev, ut_app_rec_rev_eq.
  apply ut_beta_sn_var_app_rev.
Qed.

Ltac ut_normal_tac := apply ut_normal_b_prop; reflexivity.

Fact ut_id_normal n : ut_normal (ut_id n).
Proof. ut_normal_tac. Qed.

Fact ut_church_normal n m f a : ut_neutral f -> ut_normal f -> ut_normal a -> ut_normal (@ut_church n m f a).
Proof.
  intros H1 H2 H3.
  induction m; simpl; auto.
  rewrite ut_church_S_head.
  apply ut_app_normal; auto.
Qed.

Fact ut_normal_choose n (f : uterm n) : { g | f -b-> g } + { ut_normal f }.
Proof.
  induction f as [ n f IHf | n a IHa b IHb | n v ].
  
  destruct IHf as [(g & Hg) | IHf ].
  left; exists (\\ g); constructor; auto.
  right; rewrite <- ut_abs_normal; auto.
  
  destruct IHa as [ (a' & IHa) | IHa ].
  left; exists (a' @ b); constructor 2; auto.
  destruct IHb as [ (b' & IHb) | IHb ].
  left; exists (a @ b'); constructor 3; auto.
  destruct a as [ f | | ].
  left; exists (f[b]); constructor 1.
  right; apply ut_app_normal; simpl; auto.  
  right; apply ut_app_normal; simpl; auto.
  
  right; apply ut_var_normal.
Qed.

Theorem ut_beta_normal_form n (f : uterm n) : ut_beta_sn f -> { g | f -b>> g /\ ut_normal g }.
Proof.
  induction 1 as [ f Hf IHf ].
  destruct (ut_normal_choose f) as [ (g & Hg) | H ].
  destruct (IHf _ Hg) as (g' & ? & ?).
  exists g'; split; auto.
  apply ut_beta_rt_trans with g; auto.
  apply ut_beta_beta_rt; auto.
  exists f; split; auto.
  apply ut_beta_rt_refl.
Qed.

Section ut_normal_rect.

  Variable P : forall n (f : uterm n), Type.

  Hypothesis HP1 : forall n (f : uterm (S n)), ut_normal f -> P f -> P (\\ f).
  Hypothesis HP2 : forall n (f a : uterm n), ut_neutral f -> ut_normal f -> ut_normal a -> P f -> P a -> P (f @ a).
  Hypothesis HP3 : forall n v, @P n (£ v).
  
  Theorem ut_normal_rect n f : @ut_normal n f -> P f.
  Proof.
    induction f as [ n f Hf | n f Hf a Ha | n v ]; intros H.
    apply ut_abs_normal in H; apply HP1; auto.
    apply ut_normal_app in H; apply HP2; tauto.
    auto.
  Qed.
  
End ut_normal_rect.




