(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec.
Require Import rec_alg.

Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_normal uterm_head uterm_solvable uterm_head_sem.
Require Import uterm_bool uterm_nat uterm_value.
Require Import uterm_apply uterm_fixpoint uterm_specialize.

Set Implicit Arguments.

(* st_repr : strongly represents
   ut_repr : represents
*)

(* f is strongly represented by tf iff 
   1/ tf is a functional value
   2/ if tf applied to the church numerals of v is solvable that f v has a value
   3/ if f v has a value x then tf applied to those church numerals of v reduces to the church numeral of x
*) 

Definition st_repr n k (f : natfun k) tf := 
                 @ut_fun_value n k tf
    /\ forall v, (ut_solvable (tf @* map (fun p => ut_nat n (vec_pos v p)) (pos_list k)) -> exists a, f v a) 
              /\ (forall a, f v a -> tf @* map (fun p => ut_nat n (vec_pos v p)) (pos_list k) ~b ut_nat n a).

Fact st_repr_eq n k f g t : (forall v x, f v x <-> g v x) -> @st_repr n k f t -> st_repr g t.
Proof.
  intros H [H1 H2]; split; auto.
  intros v; specialize (H2 v); destruct H2 as (H2 & H3); split.
  intros H4; apply H2 in H4; destruct H4 as (a & ?); exists a; apply H; auto.
  intros a Ha; apply H in Ha; revert Ha; apply H3.
Qed.

(* f is represented by tf : f v has value x iff tf applied to the church numerals of v reduces to that of x *)

Definition ut_repr n k (f : natfun k) tf := 
  forall v a, f v a <-> tf @* map (fun p => ut_nat n (vec_pos v p)) (pos_list k) ~b ut_nat _ a.

(* strongly is stronger than ... *)

Fact st_repr_ut_repr n k f tf : @st_repr n k f tf -> ut_repr f tf.
Proof.
  intros ( _ & H ) v; split.
  intros Ha.
  generalize (H v); clear H; intros [ H1 H2 ].
  apply H2, Ha.
  intros H0.
  generalize (H v); clear H; intros [ H1 H2 ].
  destruct H1 as (b & H1).
  apply ut_beta_eq_solvable in H0.
  apply H0, ut_nat_solvable.
  specialize (H2 _ H1).
  cutrewrite (a=b); auto.
  apply ut_nat_inj with n.
  apply ut_beta_eq_normal; try apply ut_nat_normal.
  apply ut_beta_eq_trans with (2 := H2).
  apply ut_beta_eq_sym; auto.
Qed.

Fact ut_repr_st_repr n k f tf : @ut_repr n k f tf -> ut_fun_value k tf -> st_repr f tf.
Proof.
  intros H1 H2; split; auto.
  intros v; split.
  intros Hv.
  apply H2 in Hv.
  destruct Hv as (i & Hi); exists i.
  apply H1; auto.
  intro; auto.
  intros a Ha.
  apply H1; auto.
Qed.

Hint Resolve ut_nat_normal ut_nat_solvable ut_normal_solvable ut_abs_rec_normal ut_var_normal.

Fact st_repr_cst n c : st_repr (s_cst c) (ut_nat n c).
Proof.
  split.
  intros ? ?; simpl; auto.
  split.
  intros _; simpl; auto.
  exists c; red; auto.  
  unfold s_cst; intros a ?; subst. 
  constructor 2.
Qed.

Fact st_repr_succ n : st_repr s_succ (ut_succ n).
Proof.
  split.
  
  intros v Hv H.
  simpl in H.
  apply ut_succ_solvable_app, Hv in H.
  destruct H as (i & Hi).
  exists (S i).
  generalize (ut_succ_prop n i); intros H.
  apply ut_beta_rt_eq in H.
  apply ut_beta_eq_trans with (2 := H).
  simpl.
  apply ut_beta_eq_app; auto.
  constructor 2.
  
  intros v; split.
  exists (S (vec_head v)); reflexivity.  
  intros a Ha; red in Ha; subst.
  simpl.
  apply ut_beta_rt_eq.
  rewrite vec_pos0.
  apply ut_succ_prop.
Qed.

Fact st_repr_pred n : st_repr s_pred (ut_pred n).
Proof.
  split.

  intros v Hv; simpl.
  specialize (Hv pos0); revert Hv.
  generalize (v pos0); clear v.
  intros u Hu H.
  apply ut_pred_solvable_app in H.
  apply Hu in H.
  destruct H as (i & Hi).
  exists (pred i).
  generalize (ut_pred_prop n i); intros H.
  apply ut_beta_rt_eq in H.
  apply ut_beta_eq_trans with (2 := H).
  apply ut_beta_eq_app; auto.
  apply ut_beta_eq_refl.
  
  intros v; unfold s_pred; split.
  exists (pred (vec_head v)); omega.
  intros ? ?; subst.
  apply ut_beta_rt_eq.
  simpl.
  apply ut_beta_rt_trans with (1 := ut_pred_prop _ _).
  apply ut_eq_beta_rt.
  f_equal.
  rewrite vec_pos0.
  omega.
Qed.

Definition ut_zero n : uterm n := \\ ut_nat _ 0.

Fact ut_zero_reduce n f : ut_zero n @ f -b>> ut_nat n 0.
Proof.
  unfold ut_zero.
  ut_beta_rt_step; ut_simpl.
  constructor 2.
Qed.

Fact st_repr_zero n : st_repr s_zero (ut_zero n).
Proof.
  split.
  
  intros v Hv _; simpl; exists 0.
  apply ut_beta_rt_eq, ut_zero_reduce.
  
  intros v; split.  
  exists 0; reflexivity.
  intros a Ha; red in Ha; subst; simpl.
  apply ut_beta_rt_eq, ut_zero_reduce.
Qed.

Definition ut_proj n m (p : pos m) : uterm n.
Proof.
  apply (ut_abs_rec n m). 
  apply ut_var.
  apply pos_lft, pos_rev, p.
Defined.

Fact ut_proj_reduce n m p s : @ut_proj n m p @* map s (pos_list m) -b>> s p.
Proof.
  unfold ut_proj.
  rewrite ut_app_rec_rev_eq.
  rewrite <- map_rev, pos_list_rev, map_map.
  apply ut_beta_rt_trans with (1 := ut_abs_rec_reduce _ _).
  simpl.
  rewrite pf_merge_lft_eq.
  rewrite pos_rev_rev.
  constructor 2.
Qed.

Fact ut_proj_solvable_inv n m p s : ut_solvable (@ut_proj n m p @* map s (pos_list m)) -> ut_solvable (s p).
Proof.
  intros H.
  generalize (ut_proj_reduce p s); intros H1.
  apply ut_beta_rt_solvable in H1.
  apply H1; auto.
Qed.

Fact st_repr_proj n m p : st_repr (@s_proj m p) (ut_proj n p).
Proof.
  split.
  
  intros v Hv H.
  apply ut_proj_solvable_inv, Hv in H.
  destruct H as (x & Hx).
  exists x.
  apply ut_beta_eq_trans with (2 := Hx).
  apply ut_beta_rt_eq.
  apply ut_proj_reduce.  
  
  intros v; split.
  
  exists (vec_pos v p); reflexivity.
  
  intros a Ha; red in Ha; subst.
  apply ut_beta_rt_eq.
  apply ut_beta_rt_trans with (1 := ut_proj_reduce _ _).
  constructor 2.
Qed.

Section st_repr_compo_rec.

  Variable (n k i : nat).
  Variable g : natfun k.
  Variable f : vec (natfun i) k.

  Variable phi : vec (uterm n) k.
  Variable psy : uterm n.

  (* X = \\x_1 \\x_2 ... \\x_k [psy, (phi_1) x_1 ... x_k, ..., (phi_i) x_1 ... x_k ] *)

  (* this is x1, x_2, ..., x_n *)

  Let ll := map (fun p => £ (pos_lft n p)) (rev (pos_list i)).

  Let ll' := map (fun p => ut_lift_rec i (vec_pos phi p) @* ll) (pos_list k).

  Definition ut_compo_rec := \\* (ut_lift_rec i psy #* ll').

  Hypothesis Hphi : forall p, st_repr (vec_pos f p) (vec_pos phi p).
  Hypothesis Hpsy : st_repr g psy.

  Fact ut_compo_rec_app_rec_beta_eq v : ut_compo_rec @* map v (pos_list i) 
                                     ~b psy #* map (fun p => vec_pos phi p @* map v (pos_list i)) (pos_list k).
  Proof.
    unfold ut_compo_rec.
    apply ut_beta_rt_eq.
    rewrite ut_app_rec_rev_eq, <- map_rev, pos_list_map_rev.
    match goal with 
      | |- \\* ?f *@ map ?s _  -b>> _ => generalize (ut_abs_rec_reduce f s)
    end; intros H1; apply ut_beta_rt_trans with (1 := H1); clear H1.
    rewrite ut_apply_rec_subst, ut_subst_lift_rec, ut_subst_id.
    unfold ll'.
    rewrite map_map.
    apply ut_eq_beta_rt; f_equal.
    apply map_ext.
    intros p.
    rewrite ut_subst_app_rec, ut_subst_lift_rec, ut_subst_id.
    f_equal.
    unfold ll.
    rewrite map_map, pos_list_map_rev.
    apply map_ext.
    intros q; simpl.
    rewrite pf_merge_lft_eq; auto.
    rewrite pos_rev_rev; auto.
  Qed.

  Fact ut_solvable_compo_rec v :           (forall p, ut_value (v p)) 
                                        -> ut_solvable (ut_compo_rec @* map v (pos_list i)) 
                                        -> exists a w, psy @* map (fun p => ut_nat _ (vec_pos w p)) (pos_list _) ~b ut_nat _ a 
                                        /\ forall p, vec_pos phi p @* map v (pos_list i) ~b ut_nat _ (vec_pos w p).
  Proof.
    intros Hv H.
    generalize (ut_compo_rec_app_rec_beta_eq v); intros H1.
    apply (ut_beta_eq_solvable H1) in H.
    
    generalize H; intros H2.
    apply ut_apply_rec_solvable in H.
    assert (forall p, ut_beta_eq_nat (vec_pos phi p @* map v (pos_list i))) as Hphi'.
      intros p.
      apply (proj1 (Forall_forall _ _) H).
      apply in_map_iff; exists p; split; auto.
      apply pos_list_prop.
    apply pos_reification in Hphi'.
    destruct Hphi' as (w & Hw).

    assert (Forall (@ut_beta_eq_nat _) (map (fun p => vec_pos phi p @* map v (pos_list i)) (pos_list k))) as H3.
      apply Forall_forall.
      intros x Hx.
      apply in_map_iff in Hx.
      destruct Hx as (q & ? & Hq); subst.
      exists (w q); apply Hw.
    generalize (ut_apply_rec_beta_eq H3 psy); clear H3; intros H3.

    apply (ut_beta_eq_solvable H3) in H2.
    apply (proj1 Hpsy (fun p : pos k => vec_pos phi p @* map v (pos_list i))) in H2.
    destruct H2 as (a & Ha).
    exists a, (vec_set_pos w); split; auto.
    
    apply ut_beta_eq_trans with (2 := Ha).
    apply ut_beta_eq_app_rec_fall, 
          Forall2_map, Forall_forall.
    intros p _; rewrite vec_pos_set; apply Hw.
    
    intro; rewrite vec_pos_set; auto.
    intros p Hp; exists (w p); auto.

    apply Forall_forall.
    intros x Hx.
    apply in_map_iff in Hx.
    destruct Hx as (y & ? & Hx); subst. 
    apply (proj1 (Hphi y) _); auto.
  Qed.

  Fact st_repr_comp : st_repr (s_comp g f) ut_compo_rec.
  Proof.
    split.

    intros v Hv H.
    destruct (ut_solvable_compo_rec _ Hv H) as (a & w & H1 & H2).
    exists a.
    apply ut_beta_eq_trans with (2 := H1).
    apply ut_beta_eq_trans with (1 := ut_compo_rec_app_rec_beta_eq _).
    match goal with |- ut_apply_rec _ ?l ~b _ => assert (Forall (@ut_beta_eq_nat _) l) as H3 end.
      apply Forall_forall.
      intros x Hx.
      apply in_map_iff in Hx.
      destruct Hx as (y & ? & _); subst.
      exists (vec_pos w y); apply H2.
    apply ut_beta_eq_trans with (1 := ut_apply_rec_beta_eq H3 _).
    apply ut_beta_eq_sym, ut_beta_eq_app_rec_fall, Forall2_map, Forall_forall.
    intros ? _; apply H2.

    intros v; split.
    intros H.
    
    apply ut_solvable_compo_rec in H.
    destruct H as (a & w & H1 & H2).
    exists a, w. 
    split.
    apply st_repr_ut_repr with (tf := psy); simpl; auto.
    intros p; apply st_repr_ut_repr with (tf := vec_pos phi p); simpl; auto.
    intros; apply ut_nat_value.
    
    intros a (w & H1 & H2).
    apply ut_beta_eq_trans with (1 := ut_compo_rec_app_rec_beta_eq _).
    match goal with |- ut_apply_rec _ ?l ~b _ => assert (Forall (@ut_beta_eq_nat _) l) as H3 end.
      apply Forall_forall.
      intros x Hx.
      apply in_map_iff in Hx.
      destruct Hx as (y & ? & _); subst.
      exists (vec_pos w y).
      apply st_repr_ut_repr with (1 := Hphi y); auto.
    apply ut_beta_eq_trans with (1 := ut_apply_rec_beta_eq H3 _).
    apply ut_beta_eq_trans with (2 := proj1 (st_repr_ut_repr Hpsy _ _) H1).
    apply ut_beta_eq_sym, ut_beta_eq_app_rec_fall, Forall2_map, Forall_forall.
    intros ? _; apply st_repr_ut_repr with (1 := Hphi _); auto.
  Qed.

End st_repr_compo_rec.

Section ut_fapp.

  Variables (n : nat) (f x : uterm n).

  Definition ut_fapp : uterm n := \\ (ut_lift f # £ pos0).

  Fact ut_fapp_reduce u : ut_fapp @ u -b>> f # u.
  Proof.
    unfold ut_fapp.
    ut_beta_rt_step.
    unfold ut_reduce.
    rewrite ut_apply_subst.
    apply ut_eq_beta_rt.
    f_equal.
    apply ut_reduce_lift.
  Qed.

  Fact ut_church_fapp_reduce i : ut_church (S i) ut_fapp x -b>> f # ut_church i ut_fapp x.
  Proof.
    rewrite ut_church_S_head.
    apply ut_beta_rt_trans with (1 := ut_fapp_reduce _).
    constructor 2.
  Qed.
  
  Hypothesis Hf : ut_fun_value 1 f.
  Hypothesis Hx : ut_value x.

  Fact ut_fapp_value : ut_fun_value 1 ut_fapp.
  Proof.
    intros v Hv; simpl.
    generalize (v pos0) (Hv pos0).
    clear v Hv; intros u Hu H.
    red in Hu; inst Hu.
    apply (ut_beta_rt_solvable (ut_fapp_reduce u)) in H.
    revert H; apply ut_apply_v_solvable.
    destruct Hu as (j & Hj).
    generalize (@Hf (fun _ => (ut_nat _ j))); intros H1.
    inst H1; auto.
    simpl in H1.
    apply (ut_beta_rt_solvable (ut_fapp_reduce u)) in H.
    red in H1; inst H1.
    revert H.
    apply ut_beta_eq_solvable.
    apply ut_beta_eq_sym.
    apply ut_apply_beta_eq; auto.
    destruct H1 as (z & Hz).
    exists z.
    apply ut_beta_eq_trans with (2 := Hz).
    apply ut_beta_eq_trans with (ut_apply f u).
    apply ut_beta_rt_eq, ut_fapp_reduce.
    apply ut_apply_beta_eq; auto.
  Qed.

  Fact ut_church_fapp_value i : ut_value (ut_church i ut_fapp x).
  Proof.
    apply ut_church_value; auto.
    apply ut_fapp_value.
  Qed.
  
  Fact ut_church_fapp_S i : ut_beta_eq_nat (ut_church (S i) ut_fapp x) -> ut_beta_eq_nat (ut_church i ut_fapp x).
  Proof.
    intros H; generalize (ut_beta_eq_nat_solvable H); intros H1.
    apply (ut_beta_rt_solvable (ut_church_fapp_reduce i)) in H1.
    apply ut_apply_v_solvable in H1.
    apply ut_church_fapp_value in H1; auto.
  Qed.

End ut_fapp.

Fact ut_subst_fapp n f m (s : uts_t n m) : ut_subst (@ut_fapp n f) s = ut_fapp (ut_subst f s).
Proof.
  unfold ut_fapp, ut_subst.
  fold ut_subst.
  f_equal.
  rewrite ut_apply_subst.
  f_equal.
  symmetry.
  apply ut_lift_subst_2.
Qed.

Section st_repr_iter.

  Variable (n k : nat).
  Variable f : natfun k.
  Variable g : natfun (S k).

  Variables (tf tg : uterm n).
  Hypothesis Htf : st_repr f tf.
  Hypothesis Htg : st_repr g tg.

  Let Htg' u v r : g (u##v) r <-> tg @ (ut_nat _ u) @* map (fun p => ut_nat _ (vec_pos v p)) (pos_list k) ~b ut_nat _ r.
  Proof.
    rewrite (st_repr_ut_repr Htg (u##v)).
    simpl; rewrite map_map; split; auto.
  Qed.
 
  Let tg' v := \\ (ut_lift tg @ £ pos0 @* map (fun p => ut_lift (vec_pos v p)) (pos_list k)).
  
  Let tf' v := tf @* map (fun p => vec_pos v p) (pos_list k).

  Let tf'_value v : (forall p, ut_value (vec_pos v p)) -> ut_value (tf' v).
  Proof. unfold tf'; intros; apply Htf; intro; auto. Qed.

  Let tg'_prop u v : tg' v @ u -b>> tg @ u @* map (vec_pos v) (pos_list k).
  Proof.
    unfold tg'.
    ut_beta_rt_step.
    apply ut_eq_beta_rt.
    unfold ut_reduce.
    rewrite ut_subst_app_rec.
    simpl.
    rewrite map_map.
    f_equal.
    f_equal.
    apply ut_subst_lift_first.
    apply map_ext.
    intros; apply ut_subst_lift_first.
  Qed.

  Let tg'_value v : (forall p, ut_value (vec_pos v p)) -> ut_fun_value 1 (tg' v).
  Proof.
    intros H w Hw; simpl.
    specialize (Hw pos0); revert Hw.
    generalize (w pos0); clear w.
    intros u Hu.
    
    generalize (tg'_prop u v); intros H2.    
    apply ut_beta_rt_eq, ut_beta_eq_sym in H2.
    apply ut_beta_eq_value with (1 := H2).
    generalize (proj1 Htg (vec_pos (u##v))); intros H3.
    inst H3.
    pos_auto p; simpl;auto.
    revert H3; eqgoal.
    f_equal.
    simpl; f_equal.
    rewrite map_map; apply map_ext.
    intros p; pf_auto.
  Qed.

  Let pk : pos (S k) := pos_rev pos0.
  Let psy p : pos (S k) := pos_rev (pos_nxt _ p).
  
  Definition vec_rev X n (v : vec X n) := vec_set_pos (fun p => vec_pos v (pos_rev p)).
  
  Fact vec_rev_rev X q v : vec_rev (@vec_rev X q v) = v.
  Proof.
    apply vec_pos_ext; unfold vec_rev; intro.
    repeat rewrite vec_pos_set; f_equal.
    apply pos_rev_rev.
  Qed.
  
  Let vec_pk_eq X (v : vec X _) : vec_pos v pk = vec_head (vec_rev v).
  Proof.
    rewrite <- vec_pos0; unfold vec_rev; rewrite vec_pos_set; auto.
  Qed.
  
  Let vec_set_psy_eq X (v : vec X _) : vec_set_pos (fun p => vec_pos v (psy p)) = vec_tail (vec_rev v).
  Proof.
    apply vec_pos_ext.
    unfold vec_rev; intro; repeat rewrite vec_pos_set.
    unfold psy.
    rewrite vec_pos_tail, vec_pos_set; auto.
  Qed.

  Let lpk := map psy (pos_list k).

  Definition ut_iter : uterm n.
  Proof.
    apply ut_abs_rec with (S k).
    apply ut_app.
    apply ut_app.
    apply ut_var.
    apply pos_lft, pk.
    apply ut_fapp.
    apply ut_abs.
    apply ut_app_rec.
    apply ut_app.
    apply (ut_lift_rec (S (S k)) tg).
    apply (£ pos0).
    apply (map (fun p => ut_lift (£ (pos_lft _ p))) lpk).
    apply ut_app_rec.
    apply (ut_lift_rec (S k) tf).
    apply (map (fun p => £ (pos_lft _ p)) lpk).
  Defined.

  Let ut_iter_red v : ut_iter *@ map (vec_pos v) (pos_list (S k)) 
                 -b>> vec_pos v pk @ ut_fapp (tg' (vec_set_pos (fun p => vec_pos v (psy p)))) 
                                   @ tf' (vec_set_pos (fun p => vec_pos v (psy p))).
  Proof.
    unfold ut_iter.
    apply ut_beta_rt_trans with (1 := ut_abs_rec_reduce _ _).
    apply ut_eq_beta_rt.
    unfold ut_subst; fold ut_subst.
    rewrite pf_merge_lft_eq.
    f_equal.
    f_equal.

    rewrite ut_subst_fapp.
    f_equal.
    unfold tg'.
    unfold ut_subst; fold ut_subst.
    f_equal.
    rewrite ut_subst_app_rec.
    unfold ut_subst; fold ut_subst.
    f_equal.
    f_equal.
    rewrite ut_lift_rec_subst_eq, ut_subst_comp.
    rewrite ut_lift_subst_eq.
    apply ut_subst_ext.
    intros p.
    simpl.
    unfold uts_lift.
    simpl.
    rewrite pf_merge_pos_nxt, pf_merge_rt_eq.
    auto.
    unfold lpk.
    do 2 rewrite map_map.
    apply map_ext.
    intros p.
    simpl.
    unfold uts_lift; simpl.
    f_equal.
    rewrite pf_merge_lft_eq; auto.
    rewrite vec_pos_set; auto.
    
    rewrite ut_subst_app_rec.
    unfold tf'.
    f_equal.
    rewrite ut_lift_rec_subst_eq, ut_subst_comp.
    rewrite <- (ut_subst_id tf) at 2.
    apply ut_subst_ext.
    intros p; simpl.
    rewrite pf_merge_pos_nxt, pf_merge_rt_eq; auto.
    unfold lpk.
    do 2 rewrite map_map.
    apply map_ext.
    intros p.
    simpl.
    rewrite vec_pos_set.
    unfold uts_lift; simpl.
    f_equal.
    rewrite pf_merge_lft_eq; auto.
  Qed.
  
  Let ut_iter_red' v : ut_iter *@ map (vec_pos v) (pos_list (S k)) 
                  -b>> vec_head (vec_rev v) @ ut_fapp (tg' (vec_tail (vec_rev v))) 
                                            @ tf' (vec_tail (vec_rev v)).
  Proof.
    rewrite <- vec_set_psy_eq, <- vec_pk_eq; auto.
  Qed.

  Let ut_iter_reduce u v : ut_iter @* map (vec_pos (u##v)) (pos_list (S k)) -b>> u @ ut_fapp (tg' v) @ tf' v.
  Proof.
    rewrite ut_app_rec_rev_eq, <- map_rev, pos_list_map_rev.
    replace (map (fun p => vec_pos (u##v) (pos_rev p)) (pos_list (S k))) 
       with (map (fun p => vec_pos (vec_rev (u##v)) p) (pos_list (S k))). 
    apply ut_beta_rt_trans with (1 := ut_iter_red' _).
    rewrite vec_rev_rev; unfold vec_head, vec_tail; simpl.
    constructor 2.
    apply map_ext.
    intros; unfold vec_rev; rewrite vec_pos_set; auto.
  Qed.
  
  Let ut_iter_prop i v : ut_iter @ ut_nat _ i @* map (vec_pos v) (pos_list k) ~b ut_church i (ut_fapp (tg' v)) (tf' v).
  Proof.
    apply ut_beta_rt_eq.
    generalize (ut_iter_reduce (ut_nat _ i) v).
    simpl; pf_auto; rewrite map_map; intros H.
    apply ut_beta_rt_trans with (1 := H).
    apply ut_nat_reduce.
  Qed.

  Let ut_iter_prop' i v : ut_iter @ ut_nat n (S i) @* map (vec_pos v) (pos_list k) ~b ut_apply (tg' v) (ut_iter @ ut_nat n i @* map (vec_pos v) (pos_list k)).
  Proof.
    apply ut_beta_eq_trans with (1 := ut_iter_prop _ _).
    rewrite ut_church_S_head.
    match goal with |- _ ~b ut_apply ?g ?t => apply ut_beta_eq_trans with (ut_fapp g @ t) end.
    apply ut_beta_eq_app.
    apply ut_beta_eq_refl.
    apply ut_beta_eq_sym, ut_iter_prop.
    apply ut_beta_rt_eq, ut_fapp_reduce.
  Qed.
  
  Let map_vec_pos_nxt X (v : vec X _) : map (vec_pos (vec_tail v)) (pos_list k) = map (fun p => vec_pos v (pos_nxt _ p)) (pos_list k).
  Proof.
    apply map_ext; intro; apply vec_pos_tail.
  Qed. 

  Let ut_iter_fun_value : ut_fun_value (S k) ut_iter.
  Proof.
    apply ut_fun_value_eq.
    intros v Hv H.
    generalize (ut_iter_reduce (vec_head v) (vec_tail v)) H. 
    simpl; do 2 rewrite map_map; simpl; intros H1 H2.
    rewrite map_vec_pos_nxt, <- vec_pos0 in H1.
    apply ut_beta_rt_solvable in H1.
    apply H1 in H2; clear H1.
    do 2 apply ut_app_solvable in H2.
    apply Hv in H2.
    destruct H2 as (i & Hi).
    apply ut_beta_normal_eq in Hi; auto.
    
    simpl in H; rewrite map_map in H.
    generalize (ut_iter_prop i (vec_tail v)).
    match goal with |- ut_iter @ ?x @* ?t ~b _ -> _ => assert (ut_iter @ vec_pos v pos0 @* t ~b ut_iter @ x @*t) as H1 end.
      apply ut_beta_eq_app_rec.
      apply ut_beta_rt_eq. 
      apply ut_beta_rt_rt; auto.
    intros H2.
    apply ut_beta_eq_solvable in H1.
    rewrite map_vec_pos_nxt in H1.
    apply H1 in H; clear H1.
    rewrite map_vec_pos_nxt in H2.
    apply (ut_beta_eq_solvable H2) in H.    
    apply ut_church_value in H; auto.
    destruct H as (j & Hj).
    exists j.
    apply ut_beta_eq_trans with (2 := Hj).
    apply ut_beta_eq_trans with (2 := ut_iter_prop _ _).
    rewrite map_vec_pos_nxt.
    apply ut_beta_eq_app_rec.
    apply ut_beta_rt_eq, ut_beta_rt_rt; auto.
    apply tf'_value; intros; rewrite vec_pos_tail; auto.
    apply ut_fapp_value; auto.
    apply tg'_value; intros; rewrite vec_pos_tail; auto.
  Qed.

  Let ut_iter_repr v : forall i r, s_iter_r f g i v r <-> ut_iter @ ut_nat n i @* map (fun p => ut_nat _ (vec_pos v p)) (pos_list k) ~b ut_nat _ r.
  Proof.
    assert ( (map (fun p => ut_nat n (vec_pos v p)) (pos_list k))
           = (map (vec_pos (vec_set_pos (fun p => ut_nat n (vec_pos v p)))) (pos_list k)) ) as E.
      apply map_ext; intro; rewrite vec_pos_set; auto.
    induction i as [ | i IHi ]; intros r; split; simpl.
    
    intros H.
    apply (st_repr_ut_repr Htf) in H.
    apply ut_beta_eq_trans with (2 := H).
    rewrite E.
    apply ut_beta_eq_trans with (1 := ut_iter_prop _ _).
    rewrite ut_church_O.
    unfold tf'.
    apply ut_beta_eq_refl.
 
    intros H.
    apply (st_repr_ut_repr Htf).
    apply ut_beta_eq_trans with (2 := H).
    apply ut_beta_eq_sym.
    rewrite E.
    apply ut_beta_eq_trans with (1 := ut_iter_prop _ _).
    rewrite ut_church_O.
    unfold tf'.
    apply ut_beta_eq_refl.

    intros (a & H1 & H2).
    apply IHi in H1.
    apply Htg in H2.
    rewrite E. 
    rewrite E in H1.   
    apply ut_beta_eq_trans with (1 := ut_iter_prop' i (vec_set_pos (fun p => ut_nat _ (vec_pos v p)))).
    apply ut_beta_eq_trans with (1 := ut_apply_beta_eq _ H1).
    apply ut_beta_eq_trans with (2 := H2).
    apply ut_beta_rt_eq.
    apply ut_beta_rt_trans with (1 := tg'_prop _ _).
    simpl; rewrite map_map; simpl; rewrite E; constructor 2.
    
    intros H.
    rewrite E in H.
    generalize (ut_beta_eq_trans (ut_beta_eq_sym (ut_iter_prop' i (vec_set_pos (fun p => ut_nat _ (vec_pos v p))))) H); intros H1.
    generalize H1; intros H2.
    apply ut_beta_eq_solvable, proj2 in H2; inst H2; auto.
    apply ut_apply_v_solvable in H2.
    generalize (@ut_iter_fun_value (pf_cons (ut_nat _ i) (fun p => ut_nat _ (vec_pos v p)))); intros H3.
    inst H3. intros p; invert pos p; pf_auto.
    revert H3; simpl; pf_auto; rewrite map_map; unfold pf_cons at 1; simpl; intros H3.
    rewrite E in H3.
    apply H3 in H2; clear H3.
    destruct H2 as (j & Hj).
    exists j; split.
    apply IHi; auto.
    rewrite E; auto.
    apply Htg'.
    apply ut_beta_eq_trans with (2 := H1).
    apply ut_beta_eq_sym.
    apply ut_beta_eq_trans with (1 := ut_apply_beta_eq _ Hj).
    apply ut_beta_rt_eq. 
    rewrite E; apply tg'_prop.
  Qed.

  Fact st_repr_iter : st_repr (s_iter f g) ut_iter.
  Proof.
    apply ut_repr_st_repr; auto.
    intros v a; unfold s_iter.
    rewrite ut_iter_repr.
    simpl.
    rewrite vec_pos0, map_map.
    split; eqgoal; do 2 f_equal; apply map_ext;
      intros; rewrite vec_pos_tail; auto.
  Qed.

End st_repr_iter.

Section ut_proj_bool.

  Variable (n k : nat) (phi : uterm n).

  Let f i (t : uterm i) :=  t @ \\ ut_true _ @ ut_false _.

  Definition ut_proj_bool := ut_specialize k f phi.

  Fact ut_proj_bool_breduce v t : ut_proj_bool @* map v (pos_list k) @ t 
                            -b>> phi @* t::map v (pos_list k) @ \\ ut_true _ @ ut_false _.
  Proof.
    unfold ut_proj_bool.
    apply ut_specialize_breduce with (f := f).
    intros; unfold f; auto.
    intros; unfold f; do 2 apply ut_beta_rt_lft; auto.
  Qed.

  Variable Hphi1 : ut_fun_value (S k) phi.

  Fact ut_proj_bool_fbvalue_rev : ut_fun_bvalue_rev (S k) ut_proj_bool.
  Proof.
    intros v Hv Hs.
    simpl in Hs.
    set (w := pf_cons (v (pos_fst _)) (fun p => v (pos_nxt _ (pos_rev p)))).
    assert (forall p, ut_value (w p)) as Hw.
      intros p; unfold w, pf_cons; invert pos p; simpl; apply Hv.
    generalize (Hphi1 _ Hw); intros H0.
    unfold w in H0; simpl in H0.
    unfold pf_cons in H0; simpl in H0.
    rewrite map_map in H0; simpl in H0.
    generalize (ut_beta_rt_eq (ut_proj_bool_breduce (fun p => v (pos_nxt _ (pos_rev p))) (v (pos_fst _)))); 
    intros H.
    rewrite ut_app_rev_rec_eq, map_map, <- map_rev, pos_list_map_rev in Hs.
    apply (ut_beta_eq_solvable H) in Hs.
    do 2 apply ut_app_solvable in Hs.
    apply H0 in Hs.
    destruct Hs as (i & Hi). 
    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (\\ ut_true _)) in Hi.
    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (ut_false _)) in Hi.
    simpl.
    rewrite ut_app_rev_rec_eq, map_map, <- map_rev, pos_list_map_rev.
    destruct i as [ | i ]; [ right | left ];
    apply ut_beta_eq_trans with (1 := H);
    apply ut_beta_eq_trans with (1 := Hi);
    apply ut_beta_rt_eq; 
    apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
    rewrite ut_church_O; apply ut_beta_rt_refl.
    rewrite ut_church_S_head.
    ut_beta_rt_step.
    apply ut_beta_rt_refl.
  Qed.

  Fact ut_proj_bool_fbvalue : ut_fun_bvalue (S k) ut_proj_bool.
  Proof.
     apply ut_fun_bvalue_rev_eq. 
     apply ut_proj_bool_fbvalue_rev.
  Qed.

  Fact ut_proj_bool_app_rec_bvalue (v : pos k -> uterm _) i : 
    (forall p, ut_value (v p)) -> ut_bvalue (ut_proj_bool @* map v (pos_list k) @ ut_nat n i).
  Proof.
    intros H Hi.
    rewrite ut_app_rec_rev_eq, <- map_rev, pos_list_map_rev. 
    set (s := fun p : pos (S k) => match pos_S_inv p with inl _ => ut_nat n i | inr (exist _ q _) => v (pos_rev q) end).
    destruct (@ut_proj_bool_fbvalue_rev s) as [ H1 | H1 ].
    unfold s; intros p; invert pos p; auto.
    simpl; rewrite map_map.
    unfold s; simpl.
    rewrite <- pos_list_map_rev, map_rev, <- ut_app_rec_rev_eq; auto.
    left; simpl in H1; rewrite map_map in H1; apply H1.
    right; simpl in H1; rewrite map_map in H1; apply H1.
  Qed.

  Fact ut_proj_bool_phi_O v t : 
                     phi @* t::map (fun p => ut_nat n (v p)) (pos_list k) ~b ut_nat _ 0 
                  -> ut_proj_bool @* map (fun p => ut_nat n (v p)) (pos_list k) @ t ~b ut_false _.
  Proof.
    intros Hv.

    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (\\ ut_true _)) in Hv.
    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (ut_false _)) in Hv.
    generalize (ut_proj_bool_breduce (fun p => ut_nat _ (v p)) t); intros H1.
    apply ut_beta_rt_eq in H1.
    apply ut_beta_eq_trans with (1 := H1).
    apply ut_beta_eq_trans with (1 := Hv).
    apply ut_beta_rt_eq.
    apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
    rewrite ut_church_O; apply ut_beta_rt_refl.
  Qed.

  Fact ut_proj_bool_phi_S v t i :
                     phi @* t::map (fun p => ut_nat n (v p)) (pos_list k) ~b ut_nat _ (S i) 
                  -> ut_proj_bool @* map (fun p => ut_nat n (v p)) (pos_list k) @ t ~b ut_true _.
  Proof.
    intros Hv.

    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (\\ ut_true _)) in Hv.
    apply ut_beta_eq_app with (2 := ut_beta_eq_refl (ut_false _)) in Hv.
    generalize (ut_proj_bool_breduce (fun p => ut_nat _ (v p)) t); intros H1.
    apply ut_beta_rt_eq in H1.
    apply ut_beta_eq_trans with (1 := H1).
    apply ut_beta_eq_trans with (1 := Hv).
    apply ut_beta_rt_eq.
    apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
    rewrite ut_church_S_head.
    ut_beta_rt_step; ut_simpl.    
    apply ut_beta_rt_refl.
  Qed.   

  Fact ut_proj_bool_O_phi v t :
                 ut_value t 
              -> ut_proj_bool @* map (fun p => ut_nat n (v p)) (pos_list k) @ t ~b ut_false _
              -> phi @* t::map (fun p => ut_nat n (v p)) (pos_list k) ~b ut_nat _ 0.
  Proof.
    intros Ht Hv.
    match goal with 
      |- ?x ~b _ => assert (ut_solvable (x)) as H
    end.
      generalize (ut_false_solvable n); intros H.
      apply ut_beta_eq_solvable in Hv.
      apply Hv in H.
      generalize (ut_proj_bool_breduce (fun p => ut_nat _ (v p)) t); intros H'.
      apply ut_beta_rt_solvable in H'.
      apply H' in H.
      do 2 apply ut_app_solvable in H.
      apply H.
    generalize (Hphi1 (pf_cons t (fun p => ut_nat _ (v p)))); intros H'.
    simpl in H'; rewrite map_map in H'.
    unfold pf_cons in H'; simpl in H'.
    apply H' in H.
    clear H'.
    destruct H as ([ | i ] & Hi); auto.
    apply ut_proj_bool_phi_S in Hi.
    exfalso.
    apply (@ut_true_false_beta_neq n).
    apply ut_beta_eq_trans with (2 := Hv).
    apply ut_beta_eq_sym; auto.
    intro; invert pos p; auto.
  Qed.

  Fact ut_proj_bool_S_phi v t :
                 ut_value t 
              -> ut_proj_bool @* map (fun p => ut_nat n (v p)) (pos_list k) @ t ~b ut_true _
              -> exists i, phi @* t::map (fun p => ut_nat n (v p)) (pos_list k) ~b ut_nat _ (S i).
  Proof.
    intros Ht Hv.
    match goal with 
      |- exists _, ?x ~b _ => assert (ut_solvable (x)) as H
    end.
      generalize (ut_true_solvable n); intros H.
      apply ut_beta_eq_solvable in Hv.
      apply Hv in H.
      generalize (ut_proj_bool_breduce (fun p => ut_nat _ (v p)) t); intros H'.
      apply ut_beta_rt_solvable in H'.
      apply H' in H.
      do 2 apply ut_app_solvable in H.
      apply H.
    generalize (Hphi1 (pf_cons t (fun p => ut_nat _ (v p)))); intros H'.
    simpl in H'; rewrite map_map in H'.
    unfold pf_cons in H'; simpl in H'.
    apply H' in H.
    clear H'.
    destruct H as ([ | i ] & Hi). 
    2: exists i; auto.
    apply ut_proj_bool_phi_O in Hi.
    exfalso.
    apply (@ut_true_false_beta_neq n).
    apply ut_beta_eq_trans with (2 := Hi).
    apply ut_beta_eq_sym; auto.
    intro; invert pos p; auto.
  Qed.

End ut_proj_bool.

Section st_repr_minimization.

  Variable (n k : nat).
  Variable f : natfun (S k).

  Variable phi : uterm n.
  Hypothesis Hphi : st_repr f phi.

  Let psy := ut_proj_bool k phi.

  Let psy_prop_O v i : f (i##v) 0 
                   <-> psy @* map (fun p => ut_nat n (vec_pos v p)) (pos_list k) @ ut_nat _ i ~b ut_false _.
  Proof.
    unfold psy; split; intros H.
    apply ut_proj_bool_phi_O.
    apply (st_repr_ut_repr Hphi) in H.
    simpl in H; rewrite map_map in H; apply H.
    apply (st_repr_ut_repr Hphi).
    simpl; rewrite map_map.
    apply ut_proj_bool_O_phi.
    apply Hphi.
    apply ut_nat_value.
    apply H.
  Qed.

  Let psy_prop_S v i : (exists j, f (i##v) (S j)) 
                   <-> psy @* map (fun p => ut_nat n (vec_pos v p)) (pos_list k) @ ut_nat _ i ~b ut_true _.
  Proof.
    unfold psy; split. 
    intros (j & Hj).
    apply ut_proj_bool_phi_S with (i := j).
    apply (st_repr_ut_repr Hphi) in Hj.
    simpl in Hj; rewrite map_map in Hj; apply Hj.
    intros H.
    apply ut_proj_bool_S_phi in H.
    destruct H as (j & Hj).
    exists j.
    apply (st_repr_ut_repr Hphi).
    simpl; rewrite map_map.
    apply Hj.
    apply Hphi.
    apply ut_nat_value.
  Qed.

  Let ll := map (fun p => £ (pos_lft n p)) (rev (pos_list k)).

  Definition ut_minimize := \\* (ut_find_first _ @ (ut_lift_rec k psy @* ll) @ ut_nat _ 0).

  Let ut_mini_reduce v : ut_minimize @* map v (pos_list k) -b>> ut_find_first _ @ (psy @* map v (pos_list k)) @ ut_nat _ 0.
  Proof.
    rewrite (ut_app_rec_rev_eq ut_minimize), <- map_rev, pos_list_map_rev.
    unfold ut_minimize.
    apply ut_beta_rt_trans with (1 := ut_abs_rec_reduce _ _).
    unfold ut_subst; fold ut_subst.
    rewrite ut_find_first_subst, ut_subst_app_rec, ut_subst_lift_rec, ut_subst_id, ut_nat_subst.
    apply ut_eq_beta_rt.
    do 3 f_equal.
    unfold ll.
    rewrite map_map, pos_list_map_rev.
    apply map_ext.
    intros ?; simpl; rewrite pf_merge_lft_eq, pos_rev_rev; auto.
  Qed.

  Let ut_mini_eq_nat v : ut_minimize @* map v (pos_list k) ~b ut_find_first _ @ (psy @* map v (pos_list k)) @ ut_nat _ 0.
  Proof.
    apply ut_beta_rt_eq; auto.
  Qed.

  Let ut_mini_eq_cnat v : ut_minimize @* map v (pos_list k) ~b ut_find_first _ @ (psy @* map v (pos_list k)) @ ut_cnat _ 0.
  Proof.
    generalize (ut_mini_reduce v); intros H.
    apply ut_beta_rt_eq in H.
    apply ut_beta_eq_trans with (1 := H).
    apply ut_beta_eq_refl.
  Qed.

  Fact st_repr_min : st_repr (s_min f) ut_minimize.
  Proof.
    split.
    
    intros v Hv H.
    apply (ut_beta_rt_solvable (ut_mini_reduce v)) in H.

    apply ut_find_first_wf in H.

    destruct H as (i & _ & Hi1 & Hi2).
    exists i.
    apply ut_find_first_false in Hi1.
    apply ut_beta_eq_trans with (2 := Hi1).
    apply ut_beta_eq_trans with (1 := ut_mini_eq_nat _).
    apply ut_find_first_true_seq; auto; omega.

    intro; apply ut_proj_bool_app_rec_bvalue; auto.
    apply Hphi.
    
    intros v; split.
    
    intros H.
    generalize (ut_mini_eq_nat (fun p => ut_nat _ (vec_pos v p))); intros H1.
    apply ut_beta_eq_solvable in H1.
    apply H1 in H; clear H1.
    generalize H; intros H1.
    apply ut_find_first_wf in H1.
    2: intro; apply ut_proj_bool_app_rec_bvalue; auto; apply Hphi.
    destruct H1 as (i & _ & Hi1 & Hi2).
    exists i; red; split.
    apply psy_prop_O; auto.
    intros j Hj.
    apply psy_prop_S.
    apply Hi2.
    omega.
    
    intros a (Ha1 & Ha2).
    apply psy_prop_O in Ha1.
    apply ut_beta_eq_trans with (1 := ut_mini_eq_nat _).
    apply ut_find_first_false in Ha1.
    apply ut_beta_eq_trans with (2 := Ha1).
    apply ut_find_first_true_seq.
    omega.
    intros j ( _ & Hj ).
    specialize (Ha2 _ Hj).
    apply psy_prop_S in Ha2.
    auto.
  Qed.

End st_repr_minimization.

Theorem st_repr_iteralg n k f : iteralg f -> { tf | @st_repr n _ (@recalg_rel k f) tf }.
Proof.
  intros H.
  induction H as [ c | | | | k i Hi | k i f gj Hf Hgj | k f g Hf Hg | k f Hf ] using iteralg_rect.

  exists (ut_nat _ c); apply st_repr_cst.
  exists (ut_zero _); apply st_repr_zero.
  exists (ut_succ _); apply st_repr_succ.
  exists (ut_pred _); apply st_repr_pred.

  exists (@ut_proj _ _ i); apply st_repr_proj.

  destruct Hf as (tf & Htf).
  apply pos_reif_t in Hgj; destruct Hgj as (tg & Htg).
  exists (ut_compo_rec i (vec_set_pos tg) tf); apply st_repr_comp; auto.
  intro; repeat rewrite vec_pos_set; auto.

  destruct Hf as (tf & Htf).
  destruct Hg as (tg & Htg).
  exists (ut_iter k tf tg); apply st_repr_iter; auto.

  destruct Hf as (tf & Htf).
  exists (ut_minimize k tf); apply st_repr_min; auto.
Qed.

Local Notation " '[|' f '|]' " := (@recalg_rel _ f) (at level 0).

Corollary st_repr_recalg n k (f : recalg k) : { tf : uterm n | st_repr [|f|] tf }.
Proof.
  destruct (recalg_iteralg f) as (a & H1 & H2).
  destruct (st_repr_iteralg n _ H1) as (ta & H3 & H4).
  exists ta; split; auto.
  intros v; split.
  intros Hv; destruct (proj1 (H4 v) Hv) as (u & ?); exists u.
  revert H; apply H2.
  intros u Hu; apply H4.
  revert Hu; apply H2.
Qed.

Theorem ut_repr_recalg n k (f : recalg k) : 
    { tf | forall v x, [|f|] v x <-> tf @* map (ut_nat n) (vec_list v) ~b ut_nat _ x }.
Proof.
  destruct (@st_repr_recalg n k f) as (tf & Htf).
  apply st_repr_ut_repr in Htf.
  exists tf; intros v x.
  unfold vec_list; rewrite map_map; apply Htf.
Qed.

Local Definition change_at n (p : pos n) v q := if pos_eq_dec p q then v else £ q.

Local Fact pos_eq_dec_nxt X n (p q : pos n) (x y : X) :
    (if pos_eq_dec (pos_nxt _ p) (pos_nxt _ q) then x else y)
  = (if pos_eq_dec p q then x else y).
Proof.
  destruct (pos_eq_dec p q) as [ E | C ]; subst.
  destruct (pos_eq_dec (pos_nxt _ q) (pos_nxt _ q)) as [ | C ]; auto; destruct C; auto.
  destruct (pos_eq_dec (pos_nxt _ p) (pos_nxt _ q)) as [ E | ]; auto.
  apply pos_nxt_inj in E; contradict C; auto.
Qed.

Local Lemma lemma_2_17_rec n (t : uterm n) p : 
   { t' | ut_normal t' /\ forall i, ut_subst t  (change_at p (ut_nat _ i)) 
                                 ~b ut_subst t' (change_at p (ut_nat _ i)) }.
Proof.
  revert p; induction t as [ n f | n f Hf a Ha | n v ]; intros p.

  destruct (IHf (pos_nxt _ p)) as (v & H1 & H2).
  exists (\\ v); split.
  apply ut_abs_normal; auto.
  intros i; simpl; apply ut_beta_eq_abs.
  generalize (H2 i); eqgoal.
  match goal with |- (_ _ ?x ~b _) = (_ _ ?y ~b _) 
     => assert (forall g, ut_subst g x = ut_subst g y) as E
  end.
  intros; apply ut_subst_ext.
  intros q; pos_inv q; unfold uts_lift; simpl.
  unfold change_at.
  destruct (pos_eq_dec (pos_nxt _ p) pos0) as [ E | ]; auto.
  discriminate E.
  unfold change_at.
  rewrite pos_eq_dec_nxt.
  destruct (pos_eq_dec p q) as [ E | C ]; auto.
  rewrite ut_nat_lift; auto.
  f_equal; auto.
  
  destruct (Hf p) as (f' & H1 & H2).
  destruct (Ha p) as (a' & H3 & H4).
  exists (£ p @ (\\ £ pos0) @ f' @ a'); split.
  do 3 (apply ut_app_normal; simpl; auto).
  ut_normal_tac.
  intros i.
  specialize (H2 i).
  specialize (H4 i).
  simpl ut_subst at 1.
  apply ut_beta_eq_trans with (1 := ut_beta_eq_app H2 H4).
  simpl ut_subst at 3.
  apply ut_beta_eq_app.
  2: apply ut_beta_eq_refl.
  unfold change_at at 2.
  destruct (pos_eq_dec p p) as [ _ | C ]; [ idtac | destruct C; auto ].
  change (\\ uts_lift (change_at p (ut_nat n i)) pos0) with (\\ £ pos0 : uterm n).
  apply ut_beta_eq_sym, ut_beta_rt_eq, ut_nat_id_beta_rt.

  exists (£ v).
  split.
  ut_normal_tac.
  intros i; apply ut_beta_eq_refl.
Qed.

Local Fact change_at_reduce n (t : uterm (S n)) v : ut_lift (t[ut_nat _ v]) = ut_subst t (change_at pos0 (ut_nat _ v)).
Proof.
  unfold ut_reduce.
  rewrite ut_lift_subst_1.
  apply ut_subst_ext.
  intros p; pos_inv p; unfold change_at, uts_first; simpl.
  rewrite ut_nat_lift; auto.
  unfold ut_lift; simpl; auto.
Qed.

Local Lemma lemma_2_17 n (t : uterm (S n)) : 
   { t' | ut_normal t' /\ forall i, t[ut_nat _ i] ~b t'[ut_nat _ i] }. 
Proof.
  destruct (lemma_2_17_rec t pos0) as (t' & H1 & H2).
  exists t'; split; auto.
  intros i.
  apply ut_beta_eq_inv_lift.
  do 2 rewrite change_at_reduce; auto.
Qed.

Local Lemma ut_beta_eq_common n (a b : uterm n) : a ~b b -> forall u, a ~b u <-> b ~b u.
Proof.
  intros H u; split; intros H1; apply ut_beta_eq_trans with (2 := H1); auto.
  apply ut_beta_eq_sym; auto.
Qed.

(* This is corollary 2.18 p37 *)

Corollary ut_repr_normal (f : recalg 1) : 
    { tf : uterm 0 | ut_normal tf /\ forall v x, [|f|] (v##vec_nil) x <-> tf @ ut_nat _ v ~b ut_nat _ x }.
Proof.
  destruct (ut_repr_recalg 0 f) as (tf & Htf).
  destruct (lemma_2_17 (ut_lift tf @ £ pos0)) as (t & H1 & H2).
  exists (\\ t); split.
  apply ut_abs_normal; auto.
  intros v x; rewrite Htf.
  apply ut_beta_eq_common, ut_beta_eq_sym.
  simpl.
  specialize (H2 v); simpl in H2.
  rewrite ut_reduce_app, ut_reduce_lift, ut_reduce_pos0 in H2.
  apply ut_beta_eq_sym in H2.
  apply ut_beta_eq_trans with (2 := H2).
  do 2 constructor 1.
Qed.