(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements 

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel
 
   and Lambda Calculus, Types and Models by JL Krivine

*)

Require Import List.

Require Import utils pos vec.

Require Import uterm itypes system_domega system_d.

Set Implicit Arguments.

Section semantics_for_uterms.

  Variable X : Set. (* this is the ground type for intersection types *)
  
  Notation µ := (@in_stype_0 X).
  Infix "-->" := (@in_stype_1 X) (at level 60).
  Infix "<it" := (@le_it X) (at level 70).

  Definition sem_arrow n (P Q : uterm n -> Prop) u := forall t, P t -> Q (u @ t). 

  Notation " P '##>' Q " := (@sem_arrow _ P Q) (at level 68).

  Section sem_arrow_props.

    Variable (n : nat).
    Implicit Types (P Q : uterm n -> Prop).

    Fact sem_arrow_mono P1 P2 Q1 Q2 : P2 inc1 P1 -> Q1 inc1 Q2 -> P1 ##> Q1 inc1 P2 ##> Q2.
    Proof.
      intros HP HQ ? Hu ? Ht; apply HQ, Hu, HP, Ht.
    Qed.

    Definition sem := X -> uterm n -> Prop.

    Section stype_sem.

      Variable s : sem.
     
      Let P (_ : stype X) := uterm n -> Prop.

      Let HP1 : forall x, P (µ x) := fun v => s v.
      Let HP2 : forall l c, (forall x, In x l -> P x) -> P c -> P (l --> c) :=
         fun l c Hl Hc => (fun k => forall x Hx, Hl x Hx k) ##> Hc.

      Let E : forall x, P x -> P x -> Prop := fun _ A B => forall x, A x <-> B x.

      Let HP1_ext : forall v, E (HP1 v) (HP1 v).
      Proof.
        intros ? ?; tauto.
      Qed.

      Let HP2_ext : forall l c f1 f2 k1 k2,
                        (forall x Hx, E (f1 x Hx) (f2 x Hx))
                      -> E k1 k2
                      -> E (@HP2 l c f1 k1) (@HP2 l c f2 k2).
      Proof.
        intros l c g1 g2 k1 k2 Hg Hk x; split;
        intros H a Ha; apply Hk, H; 
        intros; apply Hg; auto.
      Qed.

      Definition stype_sem : stype X -> uterm n -> Prop := stype_rect _ HP1 HP2.
      Definition itype_sem l k := forall x, In x l -> stype_sem x k.

      Fact stype_sem_var v f : stype_sem (µ v) f <-> s v f.
      Proof.
        revert f; apply stype_rect_fix0 with (1 := HP1_ext) (2 := HP2_ext).
      Qed.

      Fact stype_sem_arr l c f : 
               stype_sem (l --> c) f <-> (itype_sem l ##> stype_sem c) f.
      Proof.
        revert f; apply stype_rect_fix1 with (1 := HP1_ext) (2 := HP2_ext).
      Qed.

    End stype_sem.

    Fact le_it_sem i s t : s <it t -> itype_sem i s inc1 itype_sem i t.
    Proof.
      induction 1 as [ 
                     | s t Hs 
                     | l m a b H1 H2 
                     | l m k H1 H2
                     ] using le_it_ind; simpl; intros u Hu.

      intros ? [ ? | [] ]; subst; auto.
      intros v Hv; apply (Hs _ Hv); auto; left; auto.
      intros ? [ ? | [] ]; subst.
      specialize (Hu _ (or_introl eq_refl)).
      revert Hu.
      do 2 rewrite stype_sem_arr.
      apply sem_arrow_mono; auto.
      intros x Hx; apply H2; simpl; auto.
      intros ? [ ? | [] ]; subst; auto.
      intros x Hx; apply H2; auto.
    Qed.

  End sem_arrow_props.

  Opaque stype_sem.
  Arguments itype_sem {n} s l k /.

  Section saturated.

    Variable (n : nat).
    Implicit Types (P Q : uterm n -> Prop).
 
    Definition saturated P := forall u a l, P (u[a] @* l) -> P (\\ u @* (a::l)).

    Fact saturated_eq P Q : (forall x, P x <-> Q x) -> saturated P -> saturated Q.
    Proof.
      intros H1 H2 u a l; do 2 rewrite <- H1; apply H2.
    Qed.

    Fact saturated_right P Q : saturated Q -> saturated (P ##> Q).
    Proof.
      intros HQ u a l H p Hp.
      change (Q (\\ u @* (a::l) @* (p::nil))).
      rewrite ut_app_rec_comp.
      simpl app.
      apply HQ.
      rewrite <- ut_app_rec_comp; simpl.
      apply H, Hp.
    Qed.

    Fact saturated_inter K (l : list K) (P : forall x, In x l -> uterm n -> Prop) : 
         (forall x Hx, saturated (P x Hx)) -> saturated (fun t => forall x Hx, P x Hx t).
    Proof.
      intros HP u a m H x Hx; apply HP, H.
    Qed.

    Fact stype_sem_sat i : (forall v, saturated (i v)) 
                        -> forall t, saturated (stype_sem i t).
    Proof.
      intros H.
      induction t as [ v | l c IHl IHc ] using stype_rect.
      generalize (H v).
      apply saturated_eq; symmetry; apply stype_sem_var.
      generalize (stype_sem_arr i l c).
      intros H2; symmetry in H2.
      apply saturated_eq with (1 := H2).
      apply saturated_right; auto.
    Qed.

    Lemma krivine_3_6 (N0 N : uterm n -> Prop) :
         saturated N 
      -> N0 inc1 N
      -> N0 inc1 (fun _ => True) ##> N0
      -> N0 ##> N inc1 N
      -> forall t, N0 inc1 stype_sem (fun _ => N) t
                /\ stype_sem (fun _ => N) t inc1 N.
    Proof.
      intros H1 H2 H3 H4.
      induction t as [ x | l c IHl IHc ] using stype_rect; split.

      intro; rewrite stype_sem_var; auto.
      intros u; rewrite stype_sem_var; auto.
      
      intros u; rewrite stype_sem_arr.
      intros Hu a Ha.
      apply IHc, H3; auto.
      intros u; rewrite stype_sem_arr.
      intros Hu.
      apply H4.
      intros a Ha.
      assert (itype_sem (fun _ => N) l a) as Ha'.
        intros ? ?; apply IHl; auto.
      apply Hu in Ha'.
      revert Ha'.
      apply IHc.
    Qed.

    Definition adapted (N0 N : uterm n -> Prop) :=
         saturated N 
      /\ N0        inc1 N  ##> N0
      /\ N  ##> N0 inc1 N0 ##> N
      /\ N0 ##> N  inc1 N.
  
    Lemma krivine_3_9 (N0 N : uterm n -> Prop) :
         adapted N0 N 
      -> forall i, (forall v, N0 inc1 stype_sem i (µ v) /\ stype_sem i (µ v) inc1 N /\ saturated (stype_sem i (µ v)))
                -> (forall t, (stype_w_free_neg t -> N0 inc1 stype_sem i t) 
                           /\ (stype_w_free_pos t -> stype_sem i t inc1 N)).
    Proof.
      intros (H1 & H2 & H3 & H4) i Hi.
      induction t as [ v | l c Hl Hc ]; split.
      intros _; apply Hi.
      intros _; apply Hi.
    
      intros H; rewrite stype_w_free_neg_arr in H; destruct H as ((G1 & G2) & G3).
      generalize (fun x Hx => proj2 (Hl x Hx) (G2 _ Hx)); intros G4.
      generalize (proj1 Hc G3); intros G5.
      intros t Ht; rewrite stype_sem_arr; intros a Ha.
      apply H2 in Ht.
      apply G5, Ht.
      destruct l as [ | x ? ].
      destruct G1; auto.
      apply G4 with x; [ idtac | apply Ha ]; left; auto.
    
      intros H; rewrite stype_w_free_pos_arr in H; destruct H as (G2 & G3).
      generalize (fun x Hx => proj1 (Hl x Hx) (G2 _ Hx)); intros G4.
      generalize (proj2 Hc G3); intros G5.
      intros t; rewrite stype_sem_arr; intros Ht.
      apply H4; revert t Ht.
      apply sem_arrow_mono; auto.
      intros ? Ht ? Hx; apply (G4 _ Hx), Ht.
    Qed.

  End saturated.

  (* This is Krivine Lemma 3.5 p 44 *)

  Notation " e '|-' x ':t' t " := (@it_assign X _ e x t) (at level 70).

  Theorem system_domega_adequacy n (i : sem n) m e (f : uterm m) t s : 
        e |- f :t t 
    -> (forall p, itype_sem i (vec_pos e p) (s p))
    -> (forall v, saturated (i v))
    -> itype_sem i t (@ut_subst m f n s).
  Proof.
    intros H; revert H n i s.
    induction 1 as [ n e p t H1 
                   | n e u t H1 IH1 
                   | n e l f t H1 IH1
                   | n e f a l t H1 IH1 H2 IH2 
                   ] using it_assign_ind; intros m i s Hs Hi.

    simpl.
    intros x [ Hx | [] ]; subst x.
    apply (le_it_sem H1); auto; left; auto.

    intros x Hx.
    apply (IH1 _ Hx); auto; left; auto.

    simpl in Hs.
    simpl ut_subst.
    intros x [ E | [] ]; subst.
    rewrite stype_sem_arr.
    intros a Ha.
    apply (stype_sem_sat _ Hi _ _ _ nil); simpl. 
    unfold ut_reduce; rewrite ut_subst_comp.
    apply IH1; auto.
    intros p; pos_inv p. 
    unfold uts_first; simpl; auto.
    simpl; unfold uts_lift; simpl.
    rewrite ut_subst_lift_first; auto.
    left; auto.

    simpl ut_subst.
    intros x [ ? | [] ]; subst.
    specialize (IH1 _ _ _ Hs Hi _ (or_introl eq_refl)).
    rewrite stype_sem_arr in IH1.
    apply IH1, IH2; auto.
  Qed.

  Section Nsaturared.
  
    Variable (n : nat) (N : uterm n -> Prop).
    
    Definition Nsaturated (H : _ -> Prop) := forall u a l, N a -> H (u[a] @* l) -> H (\\ u @* (a::l)).
    
    Fact Nsaturated_eq P Q : (forall x, P x <-> Q x) -> Nsaturated P -> Nsaturated Q.
    Proof.
      intros H1 H2 u a l; do 2 rewrite <- H1; apply H2.
    Qed.

    Fact Nsaturated_right P Q : Nsaturated Q -> Nsaturated (P ##> Q).
    Proof.
      intros HQ u a l Ha H p Hp.
      change (Q (\\ u @* (a::l) @* (p::nil))).
      rewrite ut_app_rec_comp.
      simpl app.
      apply HQ; auto.
      rewrite <- ut_app_rec_comp; simpl.
      apply H, Hp.
    Qed.

    Fact Nsaturated_inter K (l : list K) (P : forall x, In x l -> uterm n -> Prop) : 
         (forall x Hx, Nsaturated (P x Hx)) -> Nsaturated (fun t => forall x Hx, P x Hx t).
    Proof.
      intros HP u a m Ha H x Hx; apply HP, H; auto.
    Qed.
    
    Fact stype_sem_Nsat i : (forall v, Nsaturated (i v)) -> forall t, Nsaturated (stype_sem i t).
    Proof.
      intros H.
      induction t as [ v | l c IHl IHc ] using stype_rect.
      generalize (H v).
      apply Nsaturated_eq; symmetry; apply stype_sem_var.
      generalize (stype_sem_arr i l c).
      intros H2; symmetry in H2.
      apply Nsaturated_eq with (1 := H2).
      apply Nsaturated_right; auto.
    Qed.

    Definition Nadapted (N0 : _ -> Prop) :=
          Nsaturated N
       /\ N0 inc1 N ##> N0
       /\ N ##> N0 inc1 N0 ##> N
       /\ N0 ##> N inc1 N.

    Lemma Krivine_3_17 N0 (i : X -> uterm n -> Prop) :
        Nadapted N0
     -> (forall x, N0 inc1 i x /\ i x inc1 N /\ Nsaturated (i x))
     ->  forall t, stype_w_free t -> N0 inc1 stype_sem i t
                                  /\ stype_sem i t inc1 N 
                                  /\ Nsaturated (stype_sem i t).
    Proof.
      intros (G1 & G2 & G3 & G4) H2.
      induction t as [ x | l t Hl Ht ]; simpl; intros H.

      split.
      intro; rewrite stype_sem_var; apply H2.
      split.
      intro; rewrite stype_sem_var; apply H2.
      generalize (proj2 (proj2 (H2 x))); apply Nsaturated_eq.
      intros; rewrite stype_sem_var; tauto.

      rewrite stype_w_free_arr in H.
      destruct H as [ F1 F2 ].
      assert (N0 inc1 itype_sem i l) as Hl1.
        intros x Hx u Hu; revert Hx; apply Hl; auto.
        apply F1; auto.
      assert (itype_sem i l inc1 N) as Hl2.
        destruct l as [ | u l ].
        destruct F1 as [ [] _ ]; auto.
        intros y Hy.
        generalize (Hy _ (or_introl eq_refl)).
        apply Hl; [ | apply F1 ]; left; auto.
      
      split.
      intros x Hx; rewrite stype_sem_arr.
      apply G2 in Hx.
      revert Hx; apply sem_arrow_mono; auto.
      apply Ht; auto.
      split.
      intros x Hx. 
      rewrite stype_sem_arr in Hx.
      apply G4; revert Hx; apply sem_arrow_mono; auto.
      apply Ht; auto.
      generalize (@Nsaturated_right (itype_sem i l) _ (proj2 (proj2 (Ht F2)))).
      apply Nsaturated_eq.
      intro; rewrite stype_sem_arr; tauto.
    Qed.

  End Nsaturared.

  (* This is Krivine 3.16 p 52 *)

  Notation " e '|-' x ':w' t " := (@it_assign_wf X _ e x t) (at level 70).
  
  Lemma system_d_adequacy n (i : sem n) m N e (f : uterm m) t s : 
        e |- f :w t
    -> (forall p, itype_sem i (vec_pos e p) (s p))
    -> (forall v, Nsaturated N (i v))
    -> (forall t, stype_w_free t -> stype_sem i t inc1 N)    
    -> itype_sem i t (@ut_subst m f n s).
  Proof.
    intros H; revert H n i N s.
    induction 1 as [ n e p t H1 H2 H3
                   | n e u t H1 H2 IH2 
                   | n e f l t H1 IH1
                   | n e f a l t H1 IH1 H2 IH2 
                   ] using it_assign_wf_ind; intros m i N s He HN Hi.

    simpl.
    intros x [ Hx | [] ]; subst x.
    apply (le_it_sem H3); auto; left; auto.

    intros x Hx.
    apply (IH2 _ Hx _ _ _ _ He HN Hi); left; auto.

    simpl in He.
    simpl ut_subst.
    generalize (ita_wf_conc H1) (ita_wf_hyps H1); intros H2 H3.
    intros x [ E | [] ]; subst.
    rewrite stype_sem_arr.
    intros a Ha.
    apply (stype_sem_Nsat _ HN _ _ _ nil).
    specialize (H3 pos0); simpl in H3.
    destruct l as [ | b l ].
    destruct (proj1 H3); auto.
    apply (Hi b), Ha.
    apply H3; left; auto.
    left; auto.
    unfold ut_reduce; rewrite ut_subst_comp.
    set (s' p := ut_subst (uts_lift s p) (uts_first a)).
    apply (IH1 _ i N s'); auto.
    intros p; pos_inv p; simpl; auto.
    replace (s' (pos_nxt _ p)) with (s p); auto.
    unfold s', uts_lift; simpl.
    rewrite ut_subst_lift_first; auto.
    left; auto.

    simpl ut_subst.
    intros x [ E | [] ]; subst.
    specialize (IH1 _ _ N _ He HN Hi _ (or_introl eq_refl)).
    rewrite stype_sem_arr in IH1.
    apply IH1, IH2 with N; auto.
  Qed.

End semantics_for_uterms.


